import collections
import inspect
from unittest import mock

import numpy as np
import pytest

from bec_lib import messages
from bec_server.device_server.tests.utils import DMMock
from bec_server.scan_server.errors import ScanAbortion
from bec_server.scan_server.scan_plugins.otf_scan import OTFScan
from bec_server.scan_server.scans import (
    Acquire,
    AddInteractiveScanPoint,
    CloseInteractiveScan,
    ContLineScan,
    DeviceRPC,
    FermatSpiralScan,
    LineScan,
    ListScan,
    MonitorScan,
    Move,
    OpenInteractiveScan,
    RequestBase,
    RoundROIScan,
    RoundScan,
    RoundScanFlySim,
    Scan,
    ScanBase,
    TimeScan,
    UpdatedMove,
    get_2D_raster_pos,
    get_fermat_spiral_pos,
    get_round_roi_scan_positions,
    get_round_scan_positions,
    unpack_scan_args,
)

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access


def test_unpack_scan_args_empty_dict():
    scan_args = {}
    expected_args = []
    assert unpack_scan_args(scan_args) == expected_args


def test_unpack_scan_args_non_dict_input():
    scan_args = [1, 2, 3]
    assert unpack_scan_args(scan_args) == scan_args


def test_unpack_scan_args_valid_input():
    scan_args = {"cmd1": [1, 2, 3], "cmd2": ["a", "b", "c"]}
    expected_args = ["cmd1", 1, 2, 3, "cmd2", "a", "b", "c"]
    assert unpack_scan_args(scan_args) == expected_args


@pytest.mark.parametrize(
    "mv_msg,reference_msg_list",
    [
        (
            messages.ScanQueueMessage(
                scan_type="mv",
                parameter={"args": {"samx": (1,), "samy": (2,)}, "kwargs": {}},
                queue="primary",
            ),
            [
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 1, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 0, "response": True},
                ),
                messages.DeviceInstructionMessage(
                    device="samy",
                    action="set",
                    parameter={"value": 2, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 1, "response": True},
                ),
            ],
        ),
        (
            messages.ScanQueueMessage(
                scan_type="mv",
                parameter={"args": {"samx": (1,), "samy": (2,), "samz": (3,)}, "kwargs": {}},
                queue="primary",
            ),
            [
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 1, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 0, "response": True},
                ),
                messages.DeviceInstructionMessage(
                    device="samy",
                    action="set",
                    parameter={"value": 2, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 1, "response": True},
                ),
                messages.DeviceInstructionMessage(
                    device="samz",
                    action="set",
                    parameter={"value": 3, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 2, "response": True},
                ),
            ],
        ),
        (
            messages.ScanQueueMessage(
                scan_type="mv", parameter={"args": {"samx": (1,)}, "kwargs": {}}, queue="primary"
            ),
            [
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 1, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 0, "response": True},
                )
            ],
        ),
    ],
)
def test_scan_move(mv_msg, reference_msg_list):
    msg_list = []
    device_manager = DMMock()
    device_manager.add_device("samx")
    device_manager.add_device("samy")
    device_manager.add_device("samz")

    def offset_mock():
        yield None

    s = Move(parameter=mv_msg.content.get("parameter"), device_manager=device_manager)
    s._set_position_offset = offset_mock
    for step in s.run():
        if step:
            msg_list.append(step)

    assert msg_list == reference_msg_list


@pytest.mark.parametrize(
    "mv_msg,reference_msg_list",
    [
        (
            messages.ScanQueueMessage(
                scan_type="umv",
                parameter={"args": {"samx": (1,), "samy": (2,)}, "kwargs": {}},
                queue="primary",
                metadata={"RID": "0bab7ee3-b384-4571-b...0fff984c05"},
            ),
            [
                messages.DeviceInstructionMessage(
                    device=None,
                    action="scan_report_instruction",
                    parameter={
                        "readback": {
                            "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                            "devices": ["samx", "samy"],
                            "start": [0, 0],
                            "end": [1.0, 2.0],
                        }
                    },
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 0,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 1.0, "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 1,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samy",
                    action="set",
                    parameter={"value": 2.0, "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 2,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="wait",
                    parameter={"type": "move", "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 3,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samy",
                    action="wait",
                    parameter={"type": "move", "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 4,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
            ],
        ),
        (
            messages.ScanQueueMessage(
                scan_type="umv",
                parameter={"args": {"samx": (1,), "samy": (2,), "samz": (3,)}, "kwargs": {}},
                queue="primary",
                metadata={"RID": "0bab7ee3-b384-4571-b...0fff984c05"},
            ),
            [
                messages.DeviceInstructionMessage(
                    device=None,
                    action="scan_report_instruction",
                    parameter={
                        "readback": {
                            "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                            "devices": ["samx", "samy", "samz"],
                            "start": [0, 0, 0],
                            "end": [1.0, 2.0, 3.0],
                        }
                    },
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 0,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 1.0, "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 1,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samy",
                    action="set",
                    parameter={"value": 2.0, "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 2,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samz",
                    action="set",
                    parameter={"value": 3.0, "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 3,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="wait",
                    parameter={"type": "move", "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 4,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samy",
                    action="wait",
                    parameter={"type": "move", "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 5,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samz",
                    action="wait",
                    parameter={"type": "move", "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 6,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
            ],
        ),
        (
            messages.ScanQueueMessage(
                scan_type="umv",
                parameter={"args": {"samx": (1,)}, "kwargs": {}},
                queue="primary",
                metadata={"RID": "0bab7ee3-b384-4571-b...0fff984c05"},
            ),
            [
                messages.DeviceInstructionMessage(
                    device=None,
                    action="scan_report_instruction",
                    parameter={
                        "readback": {
                            "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                            "devices": ["samx"],
                            "start": [0],
                            "end": [1.0],
                        }
                    },
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 0,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 1.0, "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 1,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="wait",
                    parameter={"type": "move", "wait_group": "scan_motor"},
                    metadata={
                        "readout_priority": "monitored",
                        "DIID": 2,
                        "RID": "0bab7ee3-b384-4571-b...0fff984c05",
                    },
                ),
            ],
        ),
    ],
)
def test_scan_updated_move(mv_msg, reference_msg_list):
    msg_list = []
    device_manager = DMMock()
    device_manager.add_device("samx")
    device_manager.add_device("samy")
    device_manager.add_device("samz")

    def offset_mock():
        yield None

    s = UpdatedMove(
        parameter=mv_msg.content.get("parameter"),
        device_manager=device_manager,
        metadata=mv_msg.metadata,
    )
    s._set_position_offset = offset_mock
    for step in s.run():
        if step:
            msg_list.append(step)

    assert msg_list == reference_msg_list


@pytest.mark.parametrize(
    "scan_msg,reference_scan_list",
    [
        (
            messages.ScanQueueMessage(
                scan_type="grid_scan",
                parameter={"args": {"samx": (-5, 5, 3)}, "kwargs": {}},
                queue="primary",
            ),
            [
                messages.DeviceInstructionMessage(
                    device=["samx"],
                    action="read",
                    parameter={"wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 3},
                ),
                messages.DeviceInstructionMessage(
                    device=["samx"],
                    action="wait",
                    parameter={"type": "read", "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 4},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="open_scan",
                    parameter={
                        "scan_motors": ["samx"],
                        "readout_priority": {
                            "monitored": ["samx"],
                            "baseline": [],
                            "on_request": [],
                            "async": [],
                        },
                        "num_points": 3,
                        "positions": [[-5.0], [0.0], [5.0]],
                        "scan_name": "grid_scan",
                        "scan_type": "step",
                    },
                    metadata={"readout_priority": "monitored", "DIID": 0},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="stage",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 1},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="baseline_reading",
                    parameter={},
                    metadata={"readout_priority": "baseline", "DIID": 1},
                ),
                messages.DeviceInstructionMessage(
                    **{
                        "device": "samx",
                        "action": "set",
                        "parameter": {"value": -5.0, "wait_group": "scan_motor"},
                    },
                    metadata={"readout_priority": "monitored", "DIID": 8},
                ),
                messages.DeviceInstructionMessage(
                    **{
                        "device": None,
                        "action": "wait",
                        "parameter": {
                            "type": "move",
                            "group": "scan_motor",
                            "wait_group": "scan_motor",
                        },
                    },
                    metadata={"readout_priority": "monitored", "DIID": 9},
                ),
                messages.DeviceInstructionMessage(
                    **{"device": None, "action": "pre_scan", "parameter": {}},
                    metadata={"readout_priority": "monitored", "DIID": 7},
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": -5.0, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 1},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 2},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="trigger",
                    parameter={"group": "trigger"},
                    metadata={"point_id": 0, "readout_priority": "monitored", "DIID": 3},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "trigger", "group": "trigger", "time": 0},
                    metadata={"readout_priority": "monitored", "DIID": 4},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "readout_primary"},
                    metadata={"point_id": 0, "readout_priority": "monitored", "DIID": 5},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={
                        "type": "read",
                        "group": "scan_motor",
                        "wait_group": "readout_primary",
                    },
                    metadata={"readout_priority": "monitored", "DIID": 6},
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 0.0, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 7},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 8},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 9},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="trigger",
                    parameter={"group": "trigger"},
                    metadata={"point_id": 1, "readout_priority": "monitored", "DIID": 10},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "trigger", "group": "trigger", "time": 0},
                    metadata={"readout_priority": "monitored", "DIID": 11},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "readout_primary"},
                    metadata={"point_id": 1, "readout_priority": "monitored", "DIID": 12},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={
                        "type": "read",
                        "group": "scan_motor",
                        "wait_group": "readout_primary",
                    },
                    metadata={"readout_priority": "monitored", "DIID": 13},
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 5.0, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 14},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 15},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 16},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="trigger",
                    parameter={"group": "trigger"},
                    metadata={"point_id": 2, "readout_priority": "monitored", "DIID": 17},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "trigger", "group": "trigger", "time": 0},
                    metadata={"readout_priority": "monitored", "DIID": 18},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "readout_primary"},
                    metadata={"point_id": 2, "readout_priority": "monitored", "DIID": 19},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={
                        "type": "read",
                        "group": "scan_motor",
                        "wait_group": "readout_primary",
                    },
                    metadata={"readout_priority": "monitored", "DIID": 20},
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 0.0, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 21},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 22},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 23},
                ),
                messages.DeviceInstructionMessage(
                    **{"device": None, "action": "complete", "parameter": {}},
                    metadata={"readout_priority": "monitored", "DIID": 31},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="unstage",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 24},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="close_scan",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 25},
                ),
            ],
        )
    ],
)
def test_scan_scan(scan_msg, reference_scan_list):
    device_manager = DMMock()
    device_manager.add_device("samx")
    device_manager.devices["samx"].read_buffer = {"value": 0}
    msg_list = []

    def offset_mock():
        yield None

    scan = Scan(parameter=scan_msg.content.get("parameter"), device_manager=device_manager)
    scan._set_position_offset = offset_mock
    for step in scan.run():
        if step:
            msg_list.append(step)
    scan_uid = msg_list[0].metadata.get("scan_id")
    for ii, _ in enumerate(reference_scan_list):
        if reference_scan_list[ii].metadata.get("scan_id") is not None:
            reference_scan_list[ii].metadata["scan_id"] = scan_uid
        reference_scan_list[ii].metadata["DIID"] = ii
    assert msg_list == reference_scan_list


@pytest.mark.parametrize(
    "scan_msg,reference_scan_list",
    [
        (
            messages.ScanQueueMessage(
                scan_type="fermat_scan",
                parameter={"args": {"samx": (-5, 5), "samy": (-5, 5)}, "kwargs": {"step": 3}},
                queue="primary",
            ),
            [
                (0, np.array([-1.1550884, -1.26090078])),
                (1, np.array([2.4090456, 0.21142208])),
                (2, np.array([-2.35049217, 1.80207841])),
                (3, np.array([0.59570227, -3.36772012])),
                (4, np.array([2.0522743, 3.22624707])),
                (5, np.array([-4.04502068, -1.08738572])),
                (6, np.array([4.01502502, -2.08525157])),
                (7, np.array([-1.6591442, 4.54313114])),
                (8, np.array([-1.95738438, -4.7418927])),
                (9, np.array([4.89775337, 2.29194501])),
            ],
        ),
        (
            messages.ScanQueueMessage(
                scan_type="fermat_scan",
                parameter={
                    "args": {"samx": (-5, 5), "samy": (-5, 5)},
                    "kwargs": {"step": 3, "spiral_type": 1},
                },
                queue="primary",
            ),
            [
                (0, np.array([1.1550884, 1.26090078])),
                (1, np.array([2.4090456, 0.21142208])),
                (2, np.array([2.35049217, -1.80207841])),
                (3, np.array([0.59570227, -3.36772012])),
                (4, np.array([-2.0522743, -3.22624707])),
                (5, np.array([-4.04502068, -1.08738572])),
                (6, np.array([-4.01502502, 2.08525157])),
                (7, np.array([-1.6591442, 4.54313114])),
                (8, np.array([1.95738438, 4.7418927])),
                (9, np.array([4.89775337, 2.29194501])),
            ],
        ),
    ],
)
def test_fermat_scan(scan_msg, reference_scan_list):
    device_manager = DMMock()
    device_manager.add_device("samx")
    device_manager.devices["samx"].read_buffer = {"value": 0}
    device_manager.add_device("samy")
    device_manager.devices["samy"].read_buffer = {"value": 0}
    args = unpack_scan_args(scan_msg.content.get("parameter").get("args"))
    kwargs = scan_msg.content.get("parameter").get("kwargs")
    scan = FermatSpiralScan(
        *args, parameter=scan_msg.content.get("parameter"), device_manager=device_manager, **kwargs
    )

    def offset_mock():
        yield None

    scan._set_position_offset = offset_mock
    next(scan.prepare_positions())
    # pylint: disable=protected-access
    pos = list(scan._get_position())
    assert pytest.approx(np.vstack(np.array(pos, dtype=object)[:, 1])) == np.vstack(
        np.array(reference_scan_list, dtype=object)[:, 1]
    )


@pytest.mark.parametrize(
    "scan_msg,reference_scan_list",
    [
        (
            messages.ScanQueueMessage(
                scan_type="cont_line_scan",
                parameter={"args": {"samx": (-5, 5)}, "kwargs": {"steps": 3, "exp_time": 0.1}},
                queue="primary",
            ),
            [
                messages.DeviceInstructionMessage(
                    device=["samx"],
                    action="read",
                    parameter={"wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 3},
                ),
                messages.DeviceInstructionMessage(
                    device=["samx"],
                    action="wait",
                    parameter={"type": "read", "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 4},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="open_scan",
                    parameter={
                        "scan_motors": ["samx"],
                        "readout_priority": {
                            "monitored": ["samx"],
                            "baseline": [],
                            "on_request": [],
                            "async": [],
                        },
                        "num_points": 3,
                        "positions": [[-5.0], [0.0], [5.0]],
                        "scan_name": "cont_line_scan",
                        "scan_type": "step",
                    },
                    metadata={"readout_priority": "monitored", "DIID": 0},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="stage",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 1},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="baseline_reading",
                    parameter={},
                    metadata={"readout_priority": "baseline", "DIID": 1},
                ),
                messages.DeviceInstructionMessage(
                    **{
                        "device": "samx",
                        "action": "set",
                        "parameter": {"value": -5.0, "wait_group": "scan_motor"},
                    },
                    metadata={"readout_priority": "monitored", "DIID": 5},
                ),
                messages.DeviceInstructionMessage(
                    **{
                        "device": None,
                        "action": "wait",
                        "parameter": {
                            "type": "move",
                            "group": "scan_motor",
                            "wait_group": "scan_motor",
                        },
                    },
                    metadata={"readout_priority": "monitored", "DIID": 6},
                ),
                messages.DeviceInstructionMessage(
                    **{"device": None, "action": "pre_scan", "parameter": {}},
                    metadata={"readout_priority": "monitored", "DIID": 7},
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": -105.0, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 1},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 2},
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 5.0, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 7},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="trigger",
                    parameter={"group": "trigger"},
                    metadata={"point_id": 0, "readout_priority": "monitored", "DIID": 8},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "primary"},
                    metadata={"point_id": 0, "readout_priority": "monitored", "DIID": 9},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="trigger",
                    parameter={"group": "trigger"},
                    metadata={"point_id": 1, "readout_priority": "monitored", "DIID": 10},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "primary"},
                    metadata={"point_id": 1, "readout_priority": "monitored", "DIID": 11},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="trigger",
                    parameter={"group": "trigger"},
                    metadata={"point_id": 2, "readout_priority": "monitored", "DIID": 12},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "primary"},
                    metadata={"point_id": 2, "readout_priority": "monitored", "DIID": 13},
                ),
                messages.DeviceInstructionMessage(
                    device="samx",
                    action="set",
                    parameter={"value": 0.0, "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 14},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 15},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 16},
                ),
                messages.DeviceInstructionMessage(
                    **{"device": None, "action": "complete", "parameter": {}},
                    metadata={"readout_priority": "monitored", "DIID": 31},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="unstage",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 17},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="close_scan",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 18},
                ),
            ],
        )
    ],
)
def test_cont_line_scan(scan_msg, reference_scan_list):
    device_manager = DMMock()
    device_manager.add_device("samx")
    device_manager.devices["samx"].read_buffer = {"value": 0}
    msg_list = []

    def offset_mock():
        yield None

    args = unpack_scan_args(scan_msg.content["parameter"]["args"])
    kwargs = scan_msg.content["parameter"]["kwargs"]
    scan = ContLineScan(
        *args, parameter=scan_msg.content.get("parameter"), device_manager=device_manager, **kwargs
    )
    scan._set_position_offset = offset_mock

    readback = collections.deque()
    readback.extend([{"value": -10}, {"value": -5}, {"value": 0.1}, {"value": 5}, {"value": 10}])

    def mock_readback():
        if len(readback) > 1:
            return readback.popleft()
        return readback[0]

    with mock.patch.object(scan.device_manager.devices["samx"], "readback", mock_readback):
        msg_list = [val for val in list(scan.run()) if val is not None]

        scan_uid = msg_list[0].metadata.get("scan_id")
        for ii, _ in enumerate(reference_scan_list):
            if reference_scan_list[ii].metadata.get("scan_id") is not None:
                reference_scan_list[ii].metadata["scan_id"] = scan_uid
            reference_scan_list[ii].metadata["DIID"] = ii
        assert msg_list == reference_scan_list


def test_device_rpc():
    device_manager = DMMock()
    parameter = {
        "device": "samx",
        "rpc_id": "baf7c4c0-4948-4046-8fc5-ad1e9d188c10",
        "func": "read",
        "args": [],
        "kwargs": {},
    }

    scan = DeviceRPC(parameter=parameter, device_manager=device_manager)
    scan_instructions = list(scan.run())
    assert scan_instructions == [
        messages.DeviceInstructionMessage(
            device="samx",
            action="rpc",
            parameter=parameter,
            metadata={"readout_priority": "monitored", "DIID": 0},
        )
    ]


@pytest.mark.parametrize(
    "scan_msg,reference_scan_list",
    [
        (
            messages.ScanQueueMessage(
                scan_type="acquire",
                parameter={"args": [], "kwargs": {"exp_time": 1.0}},
                queue="primary",
            ),
            [
                messages.DeviceInstructionMessage(
                    device=None,
                    action="open_scan",
                    parameter={
                        "scan_motors": [],
                        "readout_priority": {
                            "monitored": [],
                            "baseline": [],
                            "on_request": [],
                            "async": [],
                        },
                        "num_points": 1,
                        "positions": [],
                        "scan_name": "acquire",
                        "scan_type": "step",
                    },
                    metadata={"readout_priority": "monitored", "DIID": 0},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="stage",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 1},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="baseline_reading",
                    parameter={},
                    metadata={"readout_priority": "baseline", "DIID": 2},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="pre_scan",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 3},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="trigger",
                    parameter={"group": "trigger"},
                    metadata={"point_id": 0, "readout_priority": "monitored", "DIID": 3},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "trigger", "group": "trigger", "time": 1},
                    metadata={"readout_priority": "monitored", "DIID": 4},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "readout_primary"},
                    metadata={"point_id": 0, "readout_priority": "monitored", "DIID": 5},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 6},
                ),
                messages.DeviceInstructionMessage(
                    **{"device": None, "action": "complete", "parameter": {}},
                    metadata={"readout_priority": "monitored", "DIID": 31},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="unstage",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 17},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="close_scan",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 18},
                ),
            ],
        )
    ],
)
def test_acquire(scan_msg, reference_scan_list):
    device_manager = DMMock()

    scan = Acquire(exp_time=1, device_manager=device_manager)
    scan_instructions = list(scan.run())
    scan_uid = scan_instructions[0].metadata.get("scan_id")
    for ii, _ in enumerate(reference_scan_list):
        if reference_scan_list[ii].metadata.get("scan_id") is not None:
            reference_scan_list[ii].metadata["scan_id"] = scan_uid
        reference_scan_list[ii].metadata["DIID"] = ii
    assert scan_instructions == reference_scan_list


def test_pre_scan_macro():
    def pre_scan_macro(devices: dict, request: RequestBase):
        pass

    device_manager = DMMock()
    device_manager.add_device("samx")
    macros = inspect.getsource(pre_scan_macro).encode()
    scan_msg = messages.ScanQueueMessage(
        scan_type="fermat_scan",
        parameter={"args": {"samx": (-5, 5), "samy": (-5, 5)}, "kwargs": {"step": 3}},
        queue="primary",
    )
    request = FermatSpiralScan(
        device_manager=device_manager, parameter=scan_msg.content["parameter"]
    )
    with mock.patch.object(
        request.device_manager.connector,
        "lrange",
        new_callable=mock.PropertyMock,
        return_value=[messages.VariableMessage(value=macros)],
    ) as macros_mock:
        with mock.patch.object(request, "_get_func_name_from_macro", return_value="pre_scan_macro"):
            with mock.patch("builtins.eval") as eval_mock:
                request.initialize()
                eval_mock.assert_called_once_with("pre_scan_macro")


# def test_scan_report_devices():
#     device_manager = DMMock()
#     device_manager.add_device("samx")
#     parameter = {
#         "args": {"samx": (-5, 5), "samy": (-5, 5)},
#         "kwargs": {"step": 3},
#     }
#     request = RequestBase(device_manager=device_manager, parameter=parameter)
#     assert request.scan_report_devices == ["samx", "samy"]
#     request.scan_report_devices = ["samx", "samz"]
#     assert request.scan_report_devices == ["samx", "samz"]


@pytest.mark.parametrize("in_args,reference_positions", [((5, 5, 1, 1), [[1, 0], [2, 0], [-2, 0]])])
def test_round_roi_scan_positions(in_args, reference_positions):
    positions = get_round_roi_scan_positions(*in_args)
    assert np.isclose(positions, reference_positions).all()


def test_round_roi_scan():
    device_manager = DMMock()
    device_manager.add_device("samx")
    scan_msg = messages.ScanQueueMessage(
        scan_type="round_roi_scan",
        parameter={
            "args": {"samx": (10,), "samy": (10,)},
            "kwargs": {"dr": 2, "nth": 4, "exp_time": 2, "relative": True},
        },
        queue="primary",
    )
    args = unpack_scan_args(scan_msg.content["parameter"]["args"])
    kwargs = scan_msg.content["parameter"]["kwargs"]
    request = RoundROIScan(
        *args, device_manager=device_manager, parameter=scan_msg.content["parameter"], **kwargs
    )
    assert set(request.scan_report_devices) == set(["samx", "samy"])
    assert request.dr == 2
    assert request.nth == 4
    assert request.exp_time == 2
    assert request.relative is True


@pytest.mark.parametrize(
    "in_args,reference_positions", [((1, 5, 1, 1), [[0, -3], [0, -7], [0, 7]])]
)
def test_round_scan_positions(in_args, reference_positions):
    positions = get_round_scan_positions(*in_args)
    assert np.isclose(positions, reference_positions).all()


@pytest.mark.parametrize(
    "in_args,reference_positions,snaked",
    [
        (([list(range(2)), list(range(2))],), [[0, 1], [0, 0], [1, 0], [1, 1]], True),
        (
            ([list(range(2)), list(range(3))],),
            [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2]],
            False,
        ),
    ],
)
def test_raster_scan_positions(in_args, reference_positions, snaked):
    positions = get_2D_raster_pos(*in_args, snaked=snaked)
    assert np.isclose(positions, reference_positions).all()


@pytest.mark.parametrize(
    "in_args, center, reference_positions",
    [
        (
            [-2, 2, -2, 2],
            False,
            [
                [-0.38502947, -0.42030026],
                [0.8030152, 0.07047403],
                [-0.78349739, 0.6006928],
                [0.19856742, -1.12257337],
                [0.68409143, 1.07541569],
                [-1.34834023, -0.36246191],
                [1.33834167, -0.69508386],
                [-0.55304807, 1.51437705],
                [-0.65246146, -1.5806309],
                [1.63258446, 0.76398167],
                [-1.80382449, 0.565789],
                [0.99004828, -1.70839234],
                [-1.74471832, -1.22660425],
                [-1.46933912, 1.74339971],
                [1.70582397, 1.71416585],
                [1.95717083, -1.63324289],
            ],
        ),
        (
            [-1, 1, -1, 1],
            1,
            [
                [0.0, 0.0],
                [-0.38502947, -0.42030026],
                [0.8030152, 0.07047403],
                [-0.78349739, 0.6006928],
            ],
        ),
    ],
)
def test_get_fermat_spiral_pos(in_args, center, reference_positions):
    positions = get_fermat_spiral_pos(*in_args, center=center)
    assert np.isclose(positions, reference_positions).all()


def test_get_func_name_from_macro():
    def pre_scan_macro(devices: dict, request: RequestBase):
        pass

    device_manager = DMMock()
    device_manager.add_device("samx")
    macros = [inspect.getsource(pre_scan_macro).encode()]
    scan_msg = messages.ScanQueueMessage(
        scan_type="fermat_scan",
        parameter={"args": {"samx": (-5, 5), "samy": (-5, 5)}, "kwargs": {"step": 3}},
        queue="primary",
    )
    request = FermatSpiralScan(
        device_manager=device_manager, parameter=scan_msg.content["parameter"]
    )
    assert request._get_func_name_from_macro(macros[0].decode().strip()) == "pre_scan_macro"


def test_scan_report_devices():
    device_manager = DMMock()
    device_manager.add_device("samx")
    scan_msg = messages.ScanQueueMessage(
        scan_type="fermat_scan",
        parameter={"args": {"samx": (-5, 5), "samy": (-5, 5)}, "kwargs": {"step": 3}},
        queue="primary",
    )
    request = FermatSpiralScan(
        device_manager=device_manager, parameter=scan_msg.content["parameter"]
    )
    assert set(request.scan_report_devices) == set(["samx", "samy"])

    request.scan_report_devices = ["samx", "samy", "samz"]
    assert request.scan_report_devices == ["samx", "samy", "samz"]


def test_request_base_check_limits():
    class RequestBaseMock(RequestBase):
        def run(self):
            pass

    device_manager = DMMock()
    device_manager.add_device("samx")
    device_manager.add_device("samy")
    scan_msg = messages.ScanQueueMessage(
        scan_type="fermat_scan",
        parameter={"args": {"samx": (-5, 5), "samy": (-5, 5)}, "kwargs": {"step": 3}},
        queue="primary",
    )
    request = RequestBaseMock(
        device_manager=device_manager, parameter=scan_msg.content["parameter"]
    )

    assert request.scan_motors == ["samx", "samy"]
    assert request.device_manager.devices["samy"]._config["deviceConfig"].get("limits", [0, 0]) == [
        -50,
        50,
    ]
    request.device_manager.devices["samy"]._config["deviceConfig"]["limits"] = [5, -5]
    assert request.device_manager.devices["samy"]._config["deviceConfig"].get("limits", [0, 0]) == [
        5,
        -5,
    ]

    request.positions = [[-100, 30]]

    for ii, dev in enumerate(request.scan_motors):
        low_limit, high_limit = (
            request.device_manager.devices[dev]._config["deviceConfig"].get("limits", [0, 0])
        )
        for pos in request.positions:
            pos_axis = pos[ii]
            if low_limit >= high_limit:
                continue
            if not low_limit <= pos_axis <= high_limit:
                with pytest.raises(Exception) as exc_info:
                    request._check_limits()
                assert (
                    exc_info.value.args[0]
                    == f"Target position {pos} for motor {dev} is outside of range: [{low_limit},"
                    f" {high_limit}]"
                )
            else:
                request._check_limits()

    assert request.positions == [[-100, 30]]


def test_request_base_get_scan_motors():
    class RequestBaseMock(RequestBase):
        def run(self):
            pass

    device_manager = DMMock()
    device_manager.add_device("samx")
    device_manager.add_device("samz")
    scan_msg = messages.ScanQueueMessage(
        scan_type="fermat_scan",
        parameter={"args": {"samx": (-5, 5)}, "kwargs": {"step": 3}},
        queue="primary",
    )
    request = RequestBaseMock(
        device_manager=device_manager, parameter=scan_msg.content["parameter"]
    )

    assert request.scan_motors == ["samx"]
    request.caller_args = ""
    request._get_scan_motors()
    assert request.scan_motors == ["samx"]

    request.arg_bundle_size = {"bundle": 2, "min": None, "max": None}
    request.caller_args = {"samz": (-2, 2), "samy": (-1, 2)}
    request._get_scan_motors()
    assert request.scan_motors == ["samz", "samy"]

    request.caller_args = {"samx"}
    request.arg_bundle_size = {"bundle": 0, "min": None, "max": None}
    request._get_scan_motors()
    assert request.scan_motors == ["samz", "samy", "samx"]


def test_scan_base_init():
    device_manager = DMMock()
    device_manager.add_device("samx")

    class ScanBaseMock(ScanBase):
        scan_name = ""

        def _calculate_positions(self):
            pass

    scan_msg = messages.ScanQueueMessage(
        scan_type="",
        parameter={"args": {"samx": (-5, 5), "samy": (-5, 5)}, "kwargs": {"step": 3}},
        queue="primary",
    )
    with pytest.raises(ValueError) as exc_info:
        request = ScanBaseMock(
            device_manager=device_manager, parameter=scan_msg.content["parameter"]
        )
    assert exc_info.value.args[0] == "scan_name cannot be empty"


def test_scan_base_set_position_offset():
    device_manager = DMMock()
    device_manager.add_device("samx")

    scan_msg = messages.ScanQueueMessage(
        scan_type="fermat_scan",
        parameter={"args": {"samx": (-5, 5), "samy": (-5, 5)}, "kwargs": {"step": 3}},
        queue="primary",
    )
    request = FermatSpiralScan(
        device_manager=device_manager, parameter=scan_msg.content["parameter"]
    )

    assert request.positions == []
    request._set_position_offset()
    assert request.positions == []

    request.relative == True
    request._set_position_offset()

    start_pos_ref = [0, 0]
    request.positions += start_pos_ref
    assert request.positions == [0, 0]
    assert request.start_pos == start_pos_ref


def test_round_scan_fly_sim_get_scan_motors():
    device_manager = DMMock()
    device_manager.add_device("flyer_sim")
    scan_msg = messages.ScanQueueMessage(
        scan_type="round_scan_fly",
        parameter={"args": {"flyer_sim": (0, 50, 5, 3)}, "kwargs": {"realtive": True}},
        queue="primary",
    )
    request = RoundScanFlySim(
        device_manager=device_manager, parameter=scan_msg.content["parameter"]
    )

    request._get_scan_motors()
    assert request.scan_motors == []
    assert request.flyer == list(scan_msg.content["parameter"]["args"].keys())[0]


def test_round_scan_fly_sim_prepare_positions():
    device_manager = DMMock()
    device_manager.add_device("flyer_sim")
    scan_msg = messages.ScanQueueMessage(
        scan_type="round_scan_fly",
        parameter={"args": {"flyer_sim": (0, 50, 5, 3)}, "kwargs": {"realtive": True}},
        queue="primary",
    )
    request = RoundScanFlySim(
        device_manager=device_manager, parameter=scan_msg.content["parameter"]
    )
    request._calculate_positions = mock.MagicMock()
    request._check_limits = mock.MagicMock()
    pos = [1, 2, 3, 4]
    request.positions = pos

    next(request.prepare_positions())

    request._calculate_positions.assert_called_once()
    assert request.num_pos == len(pos)
    request._check_limits.assert_called_once()


@pytest.mark.parametrize(
    "in_args,reference_positions", [((1, 5, 1, 1), [[0, -3], [0, -7], [0, 7]])]
)
def test_round_scan_fly_sim_calculate_positions(in_args, reference_positions):
    device_manager = DMMock()
    device_manager.add_device("flyer_sim")
    scan_msg = messages.ScanQueueMessage(
        scan_type="round_scan_fly",
        parameter={"args": {"flyer_sim": in_args}, "kwargs": {"realtive": True}},
        queue="primary",
    )
    request = RoundScanFlySim(
        device_manager=device_manager, parameter=scan_msg.content["parameter"]
    )

    request._calculate_positions()
    assert np.isclose(request.positions, reference_positions).all()


@pytest.mark.parametrize(
    "in_args,reference_positions", [((1, 5, 1, 1), [[0, -3], [0, -7], [0, 7]])]
)
def test_round_scan_fly_sim_scan_core(in_args, reference_positions):
    device_manager = DMMock()
    device_manager.add_device("flyer_sim")
    scan_msg = messages.ScanQueueMessage(
        scan_type="round_scan_fly",
        parameter={"args": {"flyer_sim": in_args}, "kwargs": {"realtive": True}},
        queue="primary",
    )
    request = RoundScanFlySim(
        device_manager=device_manager, parameter=scan_msg.content["parameter"]
    )
    request.positions = np.array(reference_positions)

    ret = next(request.scan_core())
    assert ret == messages.DeviceInstructionMessage(
        device="flyer_sim",
        action="kickoff",
        parameter={
            "configure": {"num_pos": None, "positions": reference_positions, "exp_time": 0},
            "wait_group": "kickoff",
        },
        metadata={"readout_priority": "monitored", "DIID": 0},
    )


@pytest.mark.parametrize(
    "in_args,reference_positions",
    [
        (
            [[-3, 3], [-2, 2]],
            [
                [-3.0, -2.0],
                [-2.33333333, -1.55555556],
                [-1.66666667, -1.11111111],
                [-1.0, -0.66666667],
                [-0.33333333, -0.22222222],
                [0.33333333, 0.22222222],
                [1.0, 0.66666667],
                [1.66666667, 1.11111111],
                [2.33333333, 1.55555556],
                [3.0, 2.0],
            ],
        ),
        (
            [[-1, 1], [-1, 2]],
            [
                [-1.0, -1.0],
                [-0.77777778, -0.66666667],
                [-0.55555556, -0.33333333],
                [-0.33333333, 0.0],
                [-0.11111111, 0.33333333],
                [0.11111111, 0.66666667],
                [0.33333333, 1.0],
                [0.55555556, 1.33333333],
                [0.77777778, 1.66666667],
                [1.0, 2.0],
            ],
        ),
    ],
)
def test_line_scan_calculate_positions(in_args, reference_positions):
    device_manager = DMMock()
    scan_msg = messages.ScanQueueMessage(
        scan_type="line_scan",
        parameter={
            "args": {"samx": in_args[0], "samy": in_args[1]},
            "kwargs": {"relative": True, "steps": 10},
        },
        queue="primary",
    )
    request = LineScan(
        device_manager=device_manager,
        parameter=scan_msg.content["parameter"],
        **scan_msg.content["parameter"]["kwargs"],
    )

    request._calculate_positions()
    assert np.isclose(request.positions, reference_positions).all()


def test_list_scan_calculate_positions():
    device_manager = DMMock()
    scan_msg = messages.ScanQueueMessage(
        scan_type="list_scan",
        parameter={
            "args": {"samx": [[0, 1, 2, 3, 4]], "samy": [[0, 1, 2, 3, 4]]},
            "kwargs": {"realtive": True},
        },
        queue="primary",
    )

    request = ListScan(device_manager=device_manager, parameter=scan_msg.content["parameter"])
    request._calculate_positions()
    assert np.isclose(request.positions, [[0, 0], [1, 1], [2, 2], [3, 3], [4, 4]]).all()


def test_list_scan_raises_for_different_lengths():
    device_manager = DMMock()
    scan_msg = messages.ScanQueueMessage(
        scan_type="list_scan",
        parameter={
            "args": {"samx": [[0, 1, 2, 3, 4]], "samy": [[0, 1, 2, 3]]},
            "kwargs": {"realtive": True},
        },
        queue="primary",
    )
    with pytest.raises(ValueError):
        ListScan(device_manager=device_manager, parameter=scan_msg.content["parameter"])


@pytest.mark.parametrize(
    "scan_msg,reference_scan_list",
    [
        (
            messages.ScanQueueMessage(
                scan_type="time_scan",
                parameter={
                    "args": {},
                    "kwargs": {"points": 3, "interval": 1, "exp_time": 0.1, "relative": True},
                },
                queue="primary",
            ),
            [
                messages.DeviceInstructionMessage(
                    device=[],
                    action="read",
                    parameter={"wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 0},
                ),
                messages.DeviceInstructionMessage(
                    device=[],
                    action="wait",
                    parameter={"type": "read", "wait_group": "scan_motor"},
                    metadata={"readout_priority": "monitored", "DIID": 1},
                ),
                None,
                None,
                messages.DeviceInstructionMessage(
                    device=None,
                    action="open_scan",
                    parameter={
                        "scan_motors": [],
                        "readout_priority": {
                            "monitored": [],
                            "baseline": [],
                            "on_request": [],
                            "async": [],
                        },
                        "num_points": 3,
                        "positions": [],
                        "scan_name": "time_scan",
                        "scan_type": "step",
                    },
                    metadata={"readout_priority": "monitored", "DIID": 2},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="stage",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 3},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="baseline_reading",
                    parameter={},
                    metadata={"readout_priority": "baseline", "DIID": 4},
                ),
                messages.DeviceInstructionMessage(
                    **{"device": None, "action": "pre_scan", "parameter": {}},
                    metadata={"readout_priority": "monitored", "DIID": 5},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="trigger",
                    parameter={"group": "trigger"},
                    metadata={"readout_priority": "monitored", "DIID": 6, "point_id": 0},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "trigger", "time": 0.1, "group": "trigger"},
                    metadata={"readout_priority": "monitored", "DIID": 7},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 8, "point_id": 0},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "trigger", "time": 0.9, "group": "trigger"},
                    metadata={"readout_priority": "monitored", "DIID": 9},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 10},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="trigger",
                    parameter={"group": "trigger"},
                    metadata={"readout_priority": "monitored", "DIID": 11, "point_id": 1},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "trigger", "time": 0.1, "group": "trigger"},
                    metadata={"readout_priority": "monitored", "DIID": 12},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 13, "point_id": 1},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "trigger", "time": 0.9, "group": "trigger"},
                    metadata={"readout_priority": "monitored", "DIID": 14},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 15},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="trigger",
                    parameter={"group": "trigger"},
                    metadata={"readout_priority": "monitored", "DIID": 16, "point_id": 2},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "trigger", "time": 0.1, "group": "trigger"},
                    metadata={"readout_priority": "monitored", "DIID": 17},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 18, "point_id": 2},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "trigger", "time": 0.9, "group": "trigger"},
                    metadata={"readout_priority": "monitored", "DIID": 19},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 20},
                ),
                messages.DeviceInstructionMessage(
                    **{"device": None, "action": "complete", "parameter": {}},
                    metadata={"readout_priority": "monitored", "DIID": 21},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="unstage",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 22},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="close_scan",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 23},
                ),
            ],
        )
    ],
)
def test_time_scan(scan_msg, reference_scan_list):
    device_manager = DMMock()
    request = TimeScan(
        device_manager=device_manager,
        parameter=scan_msg.content["parameter"],
        **scan_msg.content["parameter"]["kwargs"],
    )
    scan_instructions = list(request.run())
    assert scan_instructions == reference_scan_list


@pytest.mark.parametrize(
    "scan_msg,reference_scan_list",
    [
        (
            messages.ScanQueueMessage(
                scan_type="otf_scan",
                parameter={"args": {}, "kwargs": {"e1": 700, "e2": 740, "time": 4}},
                queue="primary",
                metadata={"RID": "1234"},
            ),
            [
                None,
                None,
                None,
                messages.DeviceInstructionMessage(
                    device=None,
                    action="open_scan",
                    parameter={
                        "scan_motors": [],
                        "readout_priority": {
                            "monitored": [],
                            "baseline": [],
                            "on_request": [],
                            "async": [],
                        },
                        "num_points": 0,
                        "positions": [],
                        "scan_name": "otf_scan",
                        "scan_type": "fly",
                    },
                    metadata={"readout_priority": "monitored", "DIID": 0, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="stage",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 1, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="baseline_reading",
                    parameter={},
                    metadata={"readout_priority": "baseline", "DIID": 2, "RID": "1234"},
                ),
                None,
                messages.DeviceInstructionMessage(
                    device="mono",
                    action="set",
                    parameter={"value": 700, "wait_group": "flyer"},
                    metadata={"readout_priority": "monitored", "DIID": 3, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device=["mono"],
                    action="wait",
                    parameter={"type": "move", "wait_group": "flyer"},
                    metadata={"readout_priority": "monitored", "DIID": 4, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device="otf",
                    action="kickoff",
                    parameter={
                        "configure": {"e1": 700, "e2": 740, "time": 4},
                        "wait_group": "kickoff",
                    },
                    metadata={"readout_priority": "monitored", "DIID": 5, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device=["otf"],
                    action="wait",
                    parameter={"type": "move", "wait_group": "kickoff"},
                    metadata={"readout_priority": "monitored", "DIID": 6, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device="otf",
                    action="complete",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 7, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="read",
                    parameter={"group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 8, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 9, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="wait",
                    parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
                    metadata={"readout_priority": "monitored", "DIID": 10, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    **{"device": None, "action": "complete", "parameter": {}},
                    metadata={"readout_priority": "monitored", "DIID": 11, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="unstage",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 12, "RID": "1234"},
                ),
                messages.DeviceInstructionMessage(
                    device=None,
                    action="close_scan",
                    parameter={},
                    metadata={"readout_priority": "monitored", "DIID": 13, "RID": "1234"},
                ),
            ],
        )
    ],
)
def test_otf_scan(scan_msg, reference_scan_list):
    device_manager = DMMock()
    request = OTFScan(
        device_manager=device_manager,
        parameter=scan_msg.content["parameter"],
        metadata=scan_msg.metadata,
    )
    with mock.patch.object(request.stubs, "get_req_status", return_value=1):
        scan_instructions = list(request.run())
    assert scan_instructions == reference_scan_list


def test_monitor_scan():
    device_manager = DMMock()
    scan_msg = messages.ScanQueueMessage(
        scan_type="monitor_scan",
        parameter={"args": {"samx": [-5, 5]}, "kwargs": {"relative": True, "exp_time": 0.1}},
        queue="primary",
    )
    args = unpack_scan_args(scan_msg.content["parameter"]["args"])
    kwargs = scan_msg.content["parameter"]["kwargs"]
    request = MonitorScan(
        *args,
        device_manager=device_manager,
        parameter=scan_msg.content["parameter"],
        **scan_msg.content["parameter"]["kwargs"],
    )
    request._calculate_positions()
    assert np.isclose(request.positions, [[-5], [5]]).all()


def test_monitor_scan_run():
    device_manager = DMMock()
    scan_msg = messages.ScanQueueMessage(
        scan_type="monitor_scan",
        parameter={"args": {"samx": [-5, 5]}, "kwargs": {"relative": True, "exp_time": 0.1}},
        queue="primary",
    )
    args = unpack_scan_args(scan_msg.content["parameter"]["args"])
    kwargs = scan_msg.content["parameter"]["kwargs"]
    request = MonitorScan(
        *args,
        device_manager=device_manager,
        parameter=scan_msg.content["parameter"],
        **scan_msg.content["parameter"]["kwargs"],
    )
    with mock.patch.object(request, "_get_flyer_status") as flyer_status:
        with mock.patch.object(request, "_check_limits") as check_limits:
            with mock.patch.object(request, "_set_position_offset") as position_offset:
                flyer_status.side_effect = [
                    (None, None),
                    (None, None),
                    (None, messages.DeviceMessage(signals={"rb1": {"value": 1}})),
                    (True, None),
                ]
                ref_list = list(request.run())
                assert ref_list == [
                    messages.DeviceInstructionMessage(
                        device=["samx"],
                        action="read",
                        parameter={"wait_group": "scan_motor"},
                        metadata={"readout_priority": "monitored", "DIID": 0},
                    ),
                    messages.DeviceInstructionMessage(
                        device=["samx"],
                        action="wait",
                        parameter={"type": "read", "wait_group": "scan_motor"},
                        metadata={"readout_priority": "monitored", "DIID": 1},
                    ),
                    None,
                    messages.DeviceInstructionMessage(
                        device=None,
                        action="open_scan",
                        parameter={
                            "scan_motors": ["samx"],
                            "readout_priority": {
                                "monitored": ["samx"],
                                "baseline": [],
                                "on_request": [],
                                "async": [],
                            },
                            "num_points": 0,
                            "positions": [[-5], [5]],
                            "scan_name": "monitor_scan",
                            "scan_type": "fly",
                        },
                        metadata={"readout_priority": "monitored", "DIID": 2},
                    ),
                    messages.DeviceInstructionMessage(
                        device=None,
                        action="stage",
                        parameter={},
                        metadata={"readout_priority": "monitored", "DIID": 3},
                    ),
                    messages.DeviceInstructionMessage(
                        device=None,
                        action="baseline_reading",
                        parameter={},
                        metadata={"readout_priority": "baseline", "DIID": 4},
                    ),
                    messages.DeviceInstructionMessage(
                        device="samx",
                        action="set",
                        parameter={"value": -5, "wait_group": "scan_motor"},
                        metadata={"readout_priority": "monitored", "DIID": 5},
                    ),
                    messages.DeviceInstructionMessage(
                        device=None,
                        action="wait",
                        parameter={
                            "type": "move",
                            "group": "scan_motor",
                            "wait_group": "scan_motor",
                        },
                        metadata={"readout_priority": "monitored", "DIID": 6},
                    ),
                    messages.DeviceInstructionMessage(
                        device=None,
                        action="pre_scan",
                        parameter={},
                        metadata={"readout_priority": "monitored", "DIID": 7},
                    ),
                    messages.DeviceInstructionMessage(
                        device="samx",
                        action="set",
                        parameter={"value": -5, "wait_group": "scan_motor"},
                        metadata={"readout_priority": "monitored", "DIID": 8},
                    ),
                    messages.DeviceInstructionMessage(
                        device="samx",
                        action="wait",
                        parameter={"type": "move", "wait_group": "scan_motor"},
                        metadata={"readout_priority": "monitored", "DIID": 9},
                    ),
                    messages.DeviceInstructionMessage(
                        device="samx",
                        action="set",
                        parameter={"value": 5, "wait_group": "scan_motor"},
                        metadata={"readout_priority": "monitored", "DIID": 10, "response": True},
                    ),
                    messages.DeviceInstructionMessage(
                        device="samx",
                        action="publish_data_as_read",
                        parameter={"data": {"rb1": {"value": 1}}},
                        metadata={"readout_priority": "monitored", "DIID": 11, "point_id": 0},
                    ),
                    messages.DeviceInstructionMessage(
                        device="samx",
                        action="set",
                        parameter={"value": 0, "wait_group": "scan_motor"},
                        metadata={"readout_priority": "monitored", "DIID": 12},
                    ),
                    messages.DeviceInstructionMessage(
                        device=None,
                        action="wait",
                        parameter={
                            "type": "move",
                            "group": "scan_motor",
                            "wait_group": "scan_motor",
                        },
                        metadata={"readout_priority": "monitored", "DIID": 13},
                    ),
                    messages.DeviceInstructionMessage(
                        device=None,
                        action="wait",
                        parameter={
                            "type": "read",
                            "group": "primary",
                            "wait_group": "readout_primary",
                        },
                        metadata={"readout_priority": "monitored", "DIID": 14},
                    ),
                    messages.DeviceInstructionMessage(
                        device=None,
                        action="complete",
                        parameter={},
                        metadata={"readout_priority": "monitored", "DIID": 15},
                    ),
                    messages.DeviceInstructionMessage(
                        device=None,
                        action="unstage",
                        parameter={},
                        metadata={"readout_priority": "monitored", "DIID": 16},
                    ),
                    messages.DeviceInstructionMessage(
                        device=None,
                        action="close_scan",
                        parameter={},
                        metadata={"readout_priority": "monitored", "DIID": 17},
                    ),
                ]


def test_OpenInteractiveScan():
    device_manager = DMMock()
    scan_msg = messages.ScanQueueMessage(
        scan_type="open_interactive_scan",
        parameter={"args": {"samx": []}, "kwargs": {"relative": True, "exp_time": 0.1}},
        queue="primary",
    )
    args = unpack_scan_args(scan_msg.content["parameter"]["args"])
    kwargs = scan_msg.content["parameter"]["kwargs"]
    request = OpenInteractiveScan(
        *args,
        device_manager=device_manager,
        parameter=scan_msg.content["parameter"],
        **scan_msg.content["parameter"]["kwargs"],
    )
    ref_list = list(request.run())
    assert ref_list == [
        messages.DeviceInstructionMessage(
            device=None,
            action="open_scan_def",
            parameter={},
            metadata={"readout_priority": "monitored", "DIID": 0},
        ),
        messages.DeviceInstructionMessage(
            device=["samx"],
            action="read",
            parameter={"wait_group": "scan_motor"},
            metadata={"readout_priority": "monitored", "DIID": 1},
        ),
        messages.DeviceInstructionMessage(
            device=["samx"],
            action="wait",
            parameter={"type": "read", "wait_group": "scan_motor"},
            metadata={"readout_priority": "monitored", "DIID": 2},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="open_scan",
            parameter={
                "scan_motors": ["samx"],
                "readout_priority": {
                    "monitored": ["samx"],
                    "baseline": [],
                    "on_request": [],
                    "async": [],
                },
                "num_points": None,
                "positions": [],
                "scan_name": "open_interactive_scan",
                "scan_type": "step",
            },
            metadata={"readout_priority": "monitored", "DIID": 3},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="stage",
            parameter={},
            metadata={"readout_priority": "monitored", "DIID": 4},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="baseline_reading",
            parameter={},
            metadata={"readout_priority": "baseline", "DIID": 5},
        ),
    ]


def test_AddInteractiveScanPointn():
    device_manager = DMMock()
    scan_msg = messages.ScanQueueMessage(
        scan_type="interactive_scan_trigger",
        parameter={"args": {"samx": []}, "kwargs": {"relative": True, "exp_time": 0.1}},
        queue="primary",
    )
    args = unpack_scan_args(scan_msg.content["parameter"]["args"])
    kwargs = scan_msg.content["parameter"]["kwargs"]
    request = AddInteractiveScanPoint(
        *args,
        device_manager=device_manager,
        parameter=scan_msg.content["parameter"],
        **scan_msg.content["parameter"]["kwargs"],
    )
    ref_list = list(request.run())
    assert ref_list == [
        messages.DeviceInstructionMessage(
            device=None,
            action="open_scan",
            parameter={
                "scan_motors": ["samx"],
                "readout_priority": {
                    "monitored": ["samx"],
                    "baseline": [],
                    "on_request": [],
                    "async": [],
                },
                "num_points": None,
                "positions": [],
                "scan_name": "interactive_scan_trigger",
                "scan_type": "step",
            },
            metadata={"readout_priority": "monitored", "DIID": 0},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="trigger",
            parameter={"group": "trigger"},
            metadata={"readout_priority": "monitored", "DIID": 1, "point_id": 0},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="wait",
            parameter={"type": "trigger", "time": 0.1, "group": "trigger"},
            metadata={"readout_priority": "monitored", "DIID": 2},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="read",
            parameter={"group": "primary", "wait_group": "readout_primary"},
            metadata={"readout_priority": "monitored", "DIID": 3, "point_id": 0},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="wait",
            parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
            metadata={"readout_priority": "monitored", "DIID": 4},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="close_scan",
            parameter={},
            metadata={"readout_priority": "monitored", "DIID": 5},
        ),
    ]


def test_CloseInteractiveScan():
    device_manager = DMMock()
    device_manager.add_device("samx")
    scan_msg = messages.ScanQueueMessage(
        scan_type="close_interactive_scan",
        parameter={"args": {"samx": []}, "kwargs": {"relative": True, "exp_time": 0.1}},
        queue="primary",
    )
    args = unpack_scan_args(scan_msg.content["parameter"]["args"])
    kwargs = scan_msg.content["parameter"]["kwargs"]
    request = CloseInteractiveScan(
        *args,
        device_manager=device_manager,
        parameter=scan_msg.content["parameter"],
        **scan_msg.content["parameter"]["kwargs"],
    )
    ref_list = list(request.run())
    assert ref_list == [
        messages.DeviceInstructionMessage(
            device="samx",
            action="set",
            parameter={"value": 0, "wait_group": "scan_motor"},
            metadata={"readout_priority": "monitored", "DIID": 0},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="wait",
            parameter={"type": "move", "group": "scan_motor", "wait_group": "scan_motor"},
            metadata={"readout_priority": "monitored", "DIID": 1},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="wait",
            parameter={"type": "read", "group": "primary", "wait_group": "readout_primary"},
            metadata={"readout_priority": "monitored", "DIID": 2},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="complete",
            parameter={},
            metadata={"readout_priority": "monitored", "DIID": 3},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="unstage",
            parameter={},
            metadata={"readout_priority": "monitored", "DIID": 4},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="close_scan",
            parameter={},
            metadata={"readout_priority": "monitored", "DIID": 5},
        ),
        messages.DeviceInstructionMessage(
            device=None,
            action="close_scan_def",
            parameter={},
            metadata={"readout_priority": "monitored", "DIID": 6},
        ),
    ]


def test_RoundScan():
    device_manager = DMMock()
    scan_msg = messages.ScanQueueMessage(
        scan_type="round_scan",
        parameter={
            "args": {"samx": ["samy", 1, 2, 1, 3]},
            "kwargs": {"relative": True, "steps": 10},
        },
        queue="primary",
    )
    #     "motor_1": ScanArgType.DEVICE,
    # "motor_2": ScanArgType.DEVICE,
    # "inner_ring": ScanArgType.FLOAT,
    # "outer_ring": ScanArgType.FLOAT,
    # "number_of_rings": ScanArgType.INT,
    # "number_of_positions_in_first_ring": ScanArgType.INT,
    args = unpack_scan_args(scan_msg.content["parameter"]["args"])
    request = RoundScan(
        *args,
        device_manager=device_manager,
        parameter=scan_msg.content["parameter"],
        **scan_msg.content["parameter"]["kwargs"],
    )

    with mock.patch.object(request, "_check_limits") as check_limits:
        with mock.patch.object(request, "_set_position_offset") as position_offset:
            ref = list(request.run())
            assert len(ref) == 88



"""
This module contains the QueueItem and QueueStorage classes.
"""

from __future__ import annotations

import functools
import threading
from collections import deque
from typing import TYPE_CHECKING

from rich.console import Console
from rich.table import Table

from bec_lib import messages
from bec_lib.endpoints import MessageEndpoints
from bec_lib.utils import threadlocked

if TYPE_CHECKING:
    from bec_lib.request_items import RequestItem
    from bec_lib.scan_items import ScanItem
    from bec_lib.scan_manager import ScanManager


def update_queue(fcn):
    """Decorator to update the queue item"""

    @functools.wraps(fcn)
    def wrapper(self, *args, **kwargs):
        self._update_with_buffer()
        return fcn(self, *args, **kwargs)

    return wrapper


class QueueItem:
    # pylint: disable=too-many-arguments
    def __init__(
        self,
        scan_manager: ScanManager,
        queue_id: str,
        request_blocks: list,
        status: str,
        active_request_block: dict,
        scan_id: list[str],
        **_kwargs,
    ) -> None:
        self.scan_manager = scan_manager
        self.queue_id = queue_id
        self.request_blocks = request_blocks
        self._status = status
        self.active_request_block = active_request_block
        self.scan_ids = scan_id

    @property
    @update_queue
    def scans(self) -> list[ScanItem]:
        """get the scans items assigned to the current queue item"""
        return [
            self.scan_manager.scan_storage.find_scan_by_ID(scan_id) for scan_id in self.scan_ids
        ]

    @property
    @update_queue
    def requestIDs(self):
        return [request_block["RID"] for request_block in self.request_blocks]

    @property
    @update_queue
    def requests(self) -> list[RequestItem]:
        """get the request items assigned to the current queue item"""
        return [
            self.scan_manager.request_storage.find_request_by_ID(requestID)
            for requestID in self.requestIDs
        ]

    @property
    @update_queue
    def status(self):
        return self._status

    def _update_with_buffer(self):
        current_queue = self.scan_manager.queue_storage.current_scan_queue
        queue_info = current_queue["primary"].get("info")
        for queue_item in queue_info:
            if queue_item["queue_id"] == self.queue_id:
                self.update_queue_item(queue_item)
                return
        history = self.scan_manager.queue_storage.queue_history
        for queue_item in history:
            if queue_item.content["queue_id"] == self.queue_id:
                self.update_queue_item(queue_item.content["info"])
                return

    def update_queue_item(self, queue_item):
        """update the queue item"""
        self.request_blocks = queue_item.get("request_blocks")
        self._status = queue_item.get("status")
        self.active_request_block = queue_item.get("active_request_block")
        self.scan_ids = queue_item.get("scan_id")

    @property
    def queue_position(self) -> int | None:
        """get the current queue position"""
        current_queue = self.scan_manager.queue_storage.current_scan_queue
        for queue_group in current_queue.values():
            if not isinstance(queue_group, dict):
                continue
            for queue_position, queue in enumerate(queue_group["info"]):
                if self.queue_id == queue["queue_id"]:
                    return queue_position
        return None


class QueueStorage:
    """stores queue items"""

    def __init__(self, scan_manager: ScanManager, maxlen=50) -> None:
        self.storage: deque[QueueItem] = deque(maxlen=maxlen)
        self._lock = threading.RLock()
        self.scan_manager = scan_manager
        self.current_scan_queue = None
        self.queue_history = None

    def _update_queue_history(self):
        """get the queue history from redis"""
        self.queue_history = self.scan_manager.connector.lrange(
            MessageEndpoints.scan_queue_history(), 0, 5
        )

    def _update_current_scan_queue(self):
        """get the current scan queue from redis"""
        msg = self.scan_manager.connector.get(MessageEndpoints.scan_queue_status())
        if msg:
            self.current_scan_queue = msg.content["queue"]

    def _update_queue(self):
        self._update_current_scan_queue()
        self._update_queue_history()

    def describe_queue(self):
        """create a rich.table description of the current scan queue"""
        queue_tables = []
        self._update_queue()
        console = Console()
        for queue_name, scan_queue in self.current_scan_queue.items():
            table = Table(title=f"{queue_name} queue / {scan_queue.get('status')}")
            table.add_column("queue_id", justify="center")
            table.add_column("scan_id", justify="center")
            table.add_column("is_scan", justify="center")
            table.add_column("type", justify="center")
            table.add_column("scan_number", justify="center")
            table.add_column("IQ status", justify="center")

            for instruction_queue in scan_queue.get("info"):
                scan_msgs = [
                    msg.get("content") for msg in instruction_queue.get("request_blocks", [])
                ]
                table.add_row(
                    instruction_queue.get("queue_id"),
                    ", ".join([str(s) for s in instruction_queue.get("scan_id")]),
                    ", ".join([str(s) for s in instruction_queue.get("is_scan")]),
                    ", ".join([msg["scan_type"] for msg in scan_msgs]),
                    ", ".join([str(s) for s in instruction_queue.get("scan_number")]),
                    instruction_queue.get("status"),
                )
            with console.capture() as capture:
                console.print(table)
            queue_tables.append(capture.get())
        return queue_tables

    @threadlocked
    def update_with_status(self, queue_msg: messages.ScanQueueStatusMessage) -> None:
        """update a queue item with a new ScanQueueStatusMessage / queue message"""
        self.current_scan_queue = queue_msg.content["queue"]
        self._update_queue_history()
        queue_info = queue_msg.content["queue"]["primary"].get("info")
        for queue_item in queue_info:
            queue = self.find_queue_item_by_ID(queue_id=queue_item["queue_id"])
            if queue:
                queue.update_queue_item(queue_item)
                continue
            self.storage.append(QueueItem(scan_manager=self.scan_manager, **queue_item))

    @threadlocked
    def find_queue_item_by_ID(self, queue_id: str) -> QueueItem | None:
        """find a queue item based on its queue_id"""
        for queue_item in self.storage:
            if queue_item.queue_id == queue_id:
                return queue_item
        return None

    @threadlocked
    def find_queue_item_by_requestID(self, requestID: str) -> QueueItem | None:
        """find a queue item based on its requestID"""
        for queue_item in self.storage:
            if requestID in queue_item.requestIDs:
                return queue_item
        return None

    @threadlocked
    def find_queue_item_by_scan_id(self, scan_id: str) -> QueueItem | None:
        """find a queue item based on its scan_id"""
        for queue_item in self.storage:
            if scan_id in queue_item.scans:
                return queue_item
        return None



from typing import Protocol, runtime_checkable

from ophyd import Kind


@runtime_checkable
class BECDevice(Protocol):
    """Protocol for BEC devices with zero functionality."""

    name: str
    _destroyed: bool

    @property
    def kind(self) -> Kind:
        """kind property"""

    @kind.setter
    def kind(self, value: Kind):
        """kind setter"""

    @property
    def parent(self):
        """Property to find the parent device"""

    @property
    def root(self):
        """Property to fint the root device"""

    @property
    def hints(self) -> dict:
        """hints property"""

    @property
    def connected(self) -> bool:
        """connected property.
        Check if signals are connected

        Returns:
            bool: True if connected, False otherwise
        """

    @connected.setter
    def connected(self, value: bool):
        """connected setter"""

    def describe(self) -> dict:
        """describe method

        Includes all signals of type Kind.hinted and Kind.normal.
        Override by child class with describe method

        Returns:
            dict: Dictionary with dictionaries with signal descriptions ('source', 'dtype', 'shape')
        """

    def describe_configuration(self) -> dict:
        """describe method

        Includes all signals of type Kind.config.
        Override by child class with describe_configuration method

        Returns:
            dict: Dictionary with dictionaries with signal descriptions ('source', 'dtype', 'shape')
        """

    def read_configuration(self) -> dict:
        """read_configuration method

        Override by child class with read_configuration method

        Returns:
            dict: Dictionary with nested dictionary of signals with kind.config:
            {'signal_name' : {'value' : .., "timestamp" : ..}, ...}
        """

    def read(self) -> dict:
        """read method

        Override by child class with read method

        Returns:
            dict: Dictionary with nested dictionary of signals with kind.normal or kind.hinted:
            {'signal_name' : {'value' : .., "timestamp" : ..}, ...}
        """

    def destroy(self) -> None:
        """Destroy method"""


class BECDeviceBase:
    """Base class for BEC devices with minimum functionality.

    Device will be initiated and connected,e.g. obj.connected will be True.

    """

    def __init__(self, name: str, *args, parent=None, kind=None, **kwargs):
        self.name = name
        self._connected = True
        self._destroyed = False
        self._parent = parent
        self._kind = kind if kind else Kind.normal

    @property
    def kind(self) -> Kind:
        """Kind property, stems from ophyd."""
        return self._kind

    @kind.setter
    def kind(self, value: Kind):
        """kind setter"""
        self._kind = value

    @property
    def parent(self):
        """Property to find the parent device"""
        return self._parent

    @property
    def root(self):
        """Property to fint the root device"""
        root = self
        while True:
            if root.parent is None:
                return root
            root = root.parent

    @property
    def hints(self) -> dict:
        """hints property"""
        return {}

    @property
    def connected(self) -> bool:
        """connected property"""
        return self._connected

    @connected.setter
    def connected(self, value: bool):
        """connected setter"""
        self._connected = value

    def describe(self) -> dict:
        """describe method"""
        return {}

    def describe_configuration(self) -> dict:
        """describe_configuration method"""
        return {}

    def read(self) -> dict:
        """read method"""
        return {}

    def read_configuration(self) -> dict:
        """read_configuration method"""
        return {}

    def destroy(self) -> None:
        """destroy method"""
        self._destroyed = True
        self.connected = False



import threading
import time
from unittest import mock

import fakeredis
import pytest
import redis
from redis.client import Pipeline
from test_redis_connector import TestMessage

from bec_lib import messages
from bec_lib.endpoints import EndpointInfo, MessageEndpoints, MessageOp
from bec_lib.redis_connector import MessageObject, RedisConnector
from bec_lib.serialization import MsgpackSerialization

# pylint: disable=protected-access
# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=redefined-outer-name
# pylint: disable=unused-argument


def fake_redis_server(host, port):
    redis = fakeredis.FakeRedis()
    return redis


@pytest.fixture
def connected_connector():
    connector = RedisConnector("localhost:1", redis_cls=fake_redis_server)
    connector._redis_conn.flushall()
    try:
        yield connector
    finally:
        connector.shutdown()


TestStreamEndpoint = EndpointInfo("test", TestMessage(), MessageOp.STREAM)
TestStreamEndpoint2 = EndpointInfo("test2", TestMessage(), MessageOp.STREAM)


@pytest.mark.parametrize(
    "topics, threaded", [["topics", True], ["topics", False], [None, True], [None, False]]
)
def test_redis_connector_register_threaded(connected_connector, threaded, topics):
    connector = connected_connector
    if topics is None:
        with pytest.raises(ValueError):
            ret = connector.register(
                topics=topics, cb=lambda *args, **kwargs: ..., start_thread=threaded
            )
        return
    ret = connector.register(topics=topics, cb=lambda *args, **kwargs: ..., start_thread=threaded)
    if threaded:
        assert connector._events_listener_thread is not None


@pytest.mark.parametrize(
    "subscribed_topics, subscribed_patterns, msgs",
    [
        ["topics1", None, ["topics1"]],
        [["topics1", "topics2"], None, ["topics1", "topics2"]],
        [None, "pattern1", ["pattern1"]],
        [None, ["patt*", "top*"], ["pattern1", "topics1"]],
    ],
)
def test_redis_connector_register(
    connected_connector, subscribed_topics, subscribed_patterns, msgs
):
    connector = connected_connector
    test_msg = TestMessage()
    cb_mock = mock.Mock(spec=[])  # spec is here to remove all attributes
    if subscribed_topics:
        connector.register(
            subscribed_topics, subscribed_patterns, cb=cb_mock, start_thread=False, a=1
        )
        for msg in msgs:
            connector.send(msg, TestMessage())
            connector.poll_messages()
            msg_object = MessageObject(msg, TestMessage())
            cb_mock.assert_called_with(msg_object, a=1)
    if subscribed_patterns:
        connector.register(
            subscribed_topics, subscribed_patterns, cb=cb_mock, start_thread=False, a=1
        )
        for msg in msgs:
            connector.send(msg, TestMessage())
            connector.poll_messages()
            msg_object = MessageObject(msg, TestMessage())
            cb_mock.assert_called_with(msg_object, a=1)


def test_redis_connector_unregister(connected_connector):
    connector = connected_connector
    redisdb = connector._redis_conn

    on_msg_received = mock.Mock()
    received_event = mock.Mock(
        spec=[], side_effect=lambda msg_obj: on_msg_received(msg_obj.value.msg)
    )

    connector.register(topics=["topic1", "topic2"], cb=received_event, start_thread=False)

    connector.send("topic1", TestMessage(msg="topic1"))
    connector.poll_messages(timeout=1)
    on_msg_received.assert_called_once_with("topic1")
    connector.send("topic2", TestMessage(msg="topic2"))
    connector.poll_messages(timeout=1)
    on_msg_received.assert_called_with("topic2")

    connector.unregister("topic1", cb=received_event)

    on_msg_received.reset_mock()
    connector.send("topic1", TestMessage(msg="topic1"))
    connector.send("topic2", TestMessage(msg="topic2"))
    connector.poll_messages(timeout=1)
    on_msg_received.assert_called_once_with("topic2")

    on_msg_received.reset_mock()
    connector.unregister("topic2", cb=received_event)
    connector.send("topic1", TestMessage(msg="topic1"))
    connector.send("topic2", TestMessage(msg="topic2"))
    assert on_msg_received.call_count == 0
    assert redisdb.execute_command("PUBSUB CHANNELS") == []
    assert len(connector._topics_cb) == 0

    connector.register(topics=["topic1", "topic2"], cb=received_event, start_thread=False)

    connector.send("topic1", TestMessage(msg="topic1"))
    connector.poll_messages(timeout=1)
    on_msg_received.assert_called_once_with("topic1")
    connector.send("topic2", TestMessage(msg="topic2"))
    connector.poll_messages(timeout=1)
    on_msg_received.assert_called_with("topic2")
    connector.unregister(topics=["topic1", "topic2"])
    assert redisdb.execute_command("PUBSUB CHANNELS") == []
    assert len(connector._topics_cb) == 0


def test_redis_connector_register_identical(connected_connector):
    connector = connected_connector

    received_event1 = mock.Mock(spec=[])
    received_event2 = mock.Mock(spec=[])

    connector.register(topics="topic1", cb=received_event1, start_thread=False)
    connector.register(topics="topic1", cb=received_event1, start_thread=False)
    connector.register(topics="topic1", cb=received_event2, start_thread=False)
    connector.register(topics="topic2", cb=received_event1, start_thread=False)

    connector.send("topic1", TestMessage())
    connector.poll_messages(timeout=1)
    assert received_event1.call_count == 1
    assert received_event2.call_count == 1
    connector.send("topic2", TestMessage())
    connector.poll_messages(timeout=1)
    assert received_event1.call_count == 2


def test_redis_connector_unregister_cb_not_topic(connected_connector):
    connector = connected_connector

    topic1 = EndpointInfo("topic1", TestMessage(), MessageOp.SEND)
    topic2 = EndpointInfo("topic2", TestMessage(), MessageOp.SEND)

    received_event1 = mock.Mock(spec=[])
    received_event2 = mock.Mock(spec=[])

    connector.register(topics=topic1, cb=received_event1, start_thread=False)
    connector.register(topics=topic1, cb=received_event2, start_thread=False)
    connector.register(topics=topic2, cb=received_event1, start_thread=False)

    # normal behavior with two callbacks for the same topic
    connector.send(topic1, TestMessage())
    connector.poll_messages(timeout=1)
    assert received_event1.call_count == 1
    assert received_event2.call_count == 1

    # unregistering one callback for a topic should not remove the topic from the list
    connector.unregister(topic1, cb=received_event1)
    connector.send(topic1, TestMessage())
    connector.poll_messages(timeout=1)
    assert received_event1.call_count == 1
    assert received_event2.call_count == 2

    # unregistering the last callback for a topic should remove the topic from the list
    connector.unregister(topic1, cb=received_event2)
    connector.send(topic1, TestMessage())
    try:
        connector.poll_messages(timeout=1)
    except TimeoutError:
        pass
    assert received_event1.call_count == 1
    assert received_event2.call_count == 2


def test_redis_connector_unregister_topic_keeps_others_alive(connected_connector):
    def send_msgs_and_poll(timeout=None):
        connector.send(topic1, TestMessage())
        connector.send(topic2, TestMessage())
        time.sleep(0.1)  # give some time for messages to be received
        connector.poll_messages(timeout=timeout)

    connector = connected_connector

    topic1 = EndpointInfo("topic1", TestMessage(), MessageOp.SEND)
    topic2 = EndpointInfo("topic2", TestMessage(), MessageOp.SEND)

    received_event1 = mock.Mock(spec=[])
    received_event2 = mock.Mock(spec=[])

    connector.register(topics=topic1, cb=received_event1, start_thread=False)
    connector.register(topics=topic2, cb=received_event2, start_thread=False)

    # normal behavior with two callbacks for the two topics
    send_msgs_and_poll()
    assert received_event1.call_count == 1
    assert received_event2.call_count == 1

    # unregistering one callback/topic should leave the other topic alive
    connector.unregister(topic1, cb=received_event1)
    send_msgs_and_poll()
    assert received_event1.call_count == 1
    assert received_event2.call_count == 2

    # unregistering the last callback for a topic should remove the topic from the list
    connector.unregister(topic2, cb=received_event2)
    with pytest.raises(TimeoutError):
        send_msgs_and_poll(timeout=1)
    assert received_event1.call_count == 1
    assert received_event2.call_count == 2

    # adding a new callback for the topic should make it work again
    connector.register(topics=topic1, cb=received_event1, start_thread=False)
    send_msgs_and_poll()
    assert received_event1.call_count == 2
    assert received_event2.call_count == 2


def test_redis_register_poll_messages(connected_connector):
    connector = connected_connector
    cb_fcn_has_been_called = False

    def cb_fcn(msg, **kwargs):
        nonlocal cb_fcn_has_been_called
        cb_fcn_has_been_called = True
        assert kwargs["a"] == 1

    test_msg = TestMessage()
    connector.register("test", cb=cb_fcn, a=1, start_thread=False)
    connector._redis_conn.publish("test", MsgpackSerialization.dumps(test_msg))

    connector.poll_messages(timeout=1)

    assert cb_fcn_has_been_called

    with pytest.raises(TimeoutError):
        connector.poll_messages(timeout=0.1)


@pytest.mark.parametrize(
    "pipeline, raise_exception",
    [(None, True), (5, True), ({"a": 1}, True), (mock.MagicMock(spec=Pipeline), False)],
)
def test_redis_connector_execute_pipeline(connected_connector, pipeline, raise_exception):
    connector = connected_connector
    if raise_exception:
        with pytest.raises(TypeError):
            connector.execute_pipeline(pipeline)
    else:
        connector.execute_pipeline(pipeline)
        assert pipeline.execute.call_count == 1


def test_redis_connector_execute_pipeline_returns_list(connected_connector):
    connector = connected_connector
    pipe = connector.pipeline()
    connector.lpush(
        "test", messages.ScanMessage(point_id=5, scan_id="1234", data={"a": 1}), pipe=pipe
    )

    res = connector.execute_pipeline(pipe)
    assert isinstance(res, list) and len(res) == 1

    pipe = connector.pipeline()
    connector.lpush(
        "test", messages.ScanMessage(point_id=5, scan_id="1234", data={"a": 1}), pipe=pipe
    )
    connector.lpush(
        "test", messages.ScanMessage(point_id=5, scan_id="1234", data={"a": 1}), pipe=pipe
    )

    res = connector.execute_pipeline(pipe)
    assert isinstance(res, list) and len(res) == 2


def test_redis_connector_lpush(connected_connector):
    connector = connected_connector
    connector.lpush("test", "test_msg")
    assert connector._redis_conn.lrange("test", 0, -1) == [b"test_msg"]


def test_redis_connector_lset(connected_connector):
    connector = connected_connector
    connector.lpush("test", "test_msg")
    connector.lset("test", 0, "test_msg2")
    assert connector._redis_conn.lrange("test", 0, -1) == [b"test_msg2"]


def test_redis_connector_lset_index_out_of_range(connected_connector):
    connector = connected_connector
    connector.lpush("test", "test_msg")
    with pytest.raises(redis.exceptions.ResponseError):
        connector.lset("test", 1, "test_msg2")


def test_redis_connector_xadd(connected_connector):
    connector = connected_connector
    connector.xadd("test", {"a": 1})
    assert connector._redis_conn.xrange("test", "-", "+")[0][1] == {
        b"a": MsgpackSerialization.dumps(1)
    }

    assert connector.xread("test", count=1) == [{"a": 1}]

    assert connector.xread("test2", count=1) is None
    connector.xadd("test2", {"a": 2})
    assert connector.xread("test2", count=1) == [{"a": 2}]


def test_redis_connector_xread_repeated(connected_connector):
    connector = connected_connector
    connector.xadd("test", {"a": 1})
    connector.xadd("test", {"a": 2})
    connector.xadd("test", {"a": 3})
    assert connector.xread("test", count=2, from_start=True) == [{"a": 1}, {"a": 2}]
    assert connector.xread("test", count=2) == [{"a": 3}]
    assert connector.xread("test", count=2) is None


def test_redis_connector_xrange(connected_connector):
    connector = connected_connector
    connector.xadd("test", {"a": 1})
    connector.xadd("test", {"a": 2})
    connector.xadd("test", {"a": 3})
    assert connector.xrange("test", "-", "+", count=2) == [{"a": 1}, {"a": 2}]
    assert connector.xrange("test", "-", "+") == [{"a": 1}, {"a": 2}, {"a": 3}]

    assert connector.xrange("test2", "-", "+") is None


@pytest.mark.parametrize("endpoint", ["test", MessageEndpoints.processed_data("test")])
def test_redis_connector_get_last(connected_connector, endpoint):
    connector = connected_connector
    connector.xadd(endpoint, {"data": 1})
    connector.xadd(endpoint, {"data": 2})
    connector.xadd(endpoint, {"data": 3})
    assert connector.get_last(endpoint) == {"data": 3}
    assert connector.get_last(endpoint) == {"data": 3}
    assert connector.get_last("test2") is None
    with pytest.raises(TypeError):
        assert connector.get_last(5)
    assert list(connector.get_last(endpoint, "data", count=3)) == [1, 2, 3]
    assert list(connector.get_last(endpoint, count=4)) == [{"data": 1}, {"data": 2}, {"data": 3}]
    assert connector.get_last(endpoint, count=0) is None
    assert connector.get_last(endpoint, count=-1) is None


@pytest.mark.timeout(5)
def test_redis_connector_register_stream(connected_connector):
    connector = connected_connector
    cb_mock1 = mock.Mock(spec=[])  # spec is here to remove all attributes
    cb_mock2 = mock.Mock(spec=[])  # spec is here to remove all attributes
    connector.register(TestStreamEndpoint, cb=cb_mock1, start_thread=False, a=1)
    connector.register(TestStreamEndpoint, cb=cb_mock2, start_thread=False, a=2)
    connector.xadd("test", {"data": 1})
    connector.poll_messages()
    cb_mock1.assert_called_once_with({"data": 1}, a=1)
    cb_mock2.assert_called_once_with({"data": 1}, a=2)
    cb_mock1.reset_mock()
    cb_mock2.reset_mock()
    connector.unregister("test", cb=cb_mock1)
    connector.xadd("test", {"data": 2})
    connector.poll_messages()
    cb_mock1.assert_not_called()
    cb_mock2.assert_called_once_with({"data": 2}, a=2)
    connector.unregister("test")
    assert connector._stream_topics_subscription["test"] == []


def test_redis_connector_register_stream_identical(connected_connector):
    connector = connected_connector

    received_event1 = mock.Mock(spec=[])
    received_event2 = mock.Mock(spec=[])

    connector.register(TestStreamEndpoint, cb=received_event1, start_thread=False)
    connector.register(TestStreamEndpoint, cb=received_event1, start_thread=False)
    connector.register(TestStreamEndpoint, cb=received_event2, start_thread=False)
    connector.register(TestStreamEndpoint2, cb=received_event1, start_thread=False)
    connector.xadd("test", {"data": 1})
    connector.poll_messages(timeout=1)
    assert received_event1.call_count == 1
    assert received_event2.call_count == 1
    connector.xadd("test2", {"data": 1})
    connector.poll_messages(timeout=1)
    assert received_event1.call_count == 2

    try:
        with pytest.raises(RuntimeError):
            connector.register(
                TestStreamEndpoint2, cb=received_event1, newest_only=True, start_thread=False
            )
        connector.register(
            TestStreamEndpoint2, cb=received_event2, newest_only=True, start_thread=False
        )
        with pytest.raises(RuntimeError):
            connector.register(TestStreamEndpoint2, cb=received_event2, start_thread=False)
    finally:
        connector.unregister(TestStreamEndpoint2)


@pytest.mark.timeout(5)
@pytest.mark.parametrize(
    "endpoint",
    [
        [TestStreamEndpoint],
        [MessageEndpoints.processed_data("test"), MessageEndpoints.processed_data("test2")],
    ],
)
def test_redis_connector_register_stream_list(connected_connector, endpoint):
    connector = connected_connector
    cb_mock = mock.Mock(spec=[])  # spec is here to remove all attributes
    connector.register(endpoint, cb=cb_mock, start_thread=False, a=1)
    for ep in endpoint:
        connector.xadd(ep, {"data": 1})
        connector.poll_messages()
    assert mock.call({"data": 1}, a=1) in cb_mock.mock_calls

    for ep in endpoint:
        connector.xadd(ep, {"data": 2})
        connector.poll_messages()
    assert mock.call({"data": 2}, a=1) in cb_mock.mock_calls
    connector.unregister(endpoint)
    assert len(connector._stream_topics_subscription) == 0


@pytest.mark.timeout(5)
def test_redis_connector_register_stream_from_start(connected_connector):
    connector = connected_connector
    cb_mock1 = mock.Mock(spec=[])  # spec is here to remove all attributes
    cb_mock2 = mock.Mock(spec=[])  # spec is here to remove all attributes
    connector.xadd("test", {"data": 1})
    connector.xadd("test", {"data": 2})
    connector.register(TestStreamEndpoint, cb=cb_mock1, from_start=True, start_thread=False, a=1)
    connector.register(TestStreamEndpoint, cb=cb_mock2, start_thread=False, a=2)
    connector.poll_messages(timeout=1)
    cb_mock1.assert_has_calls([mock.call({"data": 1}, a=1), mock.call({"data": 2}, a=1)])
    cb_mock1.reset_mock()
    assert cb_mock2.call_count == 0
    connector.xadd("test", {"data": 3})
    connector.poll_messages(timeout=1)
    cb_mock1.assert_called_once_with({"data": 3}, a=1)
    cb_mock2.assert_called_once_with({"data": 3}, a=2)
    cb_mock1.reset_mock()
    connector.register(TestStreamEndpoint, cb=cb_mock1, from_start=True, start_thread=False, a=3)
    connector.poll_messages(timeout=1)
    cb_mock1.assert_has_calls(
        [mock.call({"data": 1}, a=3), mock.call({"data": 2}, a=3), mock.call({"data": 3}, a=3)]
    )


@pytest.mark.timeout(5)
def test_redis_connector_register_stream_newest_only(connected_connector):
    connector = connected_connector
    endpoint = TestStreamEndpoint
    # simulate callback taking 1s to perform its task
    cb_mock = mock.Mock(spec=[], side_effect=lambda _: time.sleep(1))

    connector.register(endpoint, cb=cb_mock, newest_only=True)
    connector.xadd(endpoint, {"data": 0})
    # from here: cb_mock will be called from another thread when new stream item is pushed
    # cb_mock.call_count will be increased before sleep time (when call has just been done)
    while cb_mock.call_count == 0:
        time.sleep(0.01)
    cb_mock.assert_called_once_with({"data": 0})
    cb_mock.reset_mock()
    # cb_mock is now sleeping...
    # Meanwhile data is published:
    connector.xadd(endpoint, {"data": 1})
    connector.xadd(endpoint, {"data": 2})
    connector.xadd(endpoint, {"data": 3})
    while cb_mock.call_count == 0:
        time.sleep(0.01)
    # cb_mock has been called for the second time ;
    # 'newest_only' means it should be called with newest data
    cb_mock.assert_called_once_with({"data": 3})

    num_threads = threading.active_count()
    connector.unregister(endpoint, cb=cb_mock)
    assert threading.active_count() == num_threads - 1


@pytest.mark.timeout(5)
def test_redis_connector_register_stream_newest_only_multiple_endpoints_same_cb(
    connected_connector,
):
    connector = connected_connector
    endpoint = [TestStreamEndpoint, TestStreamEndpoint2]
    cb_mock = mock.Mock(spec=[])  # spec is here to remove all attributes
    connector.register(endpoint, cb=cb_mock, newest_only=True, a=1)
    time.sleep(0.1)
    for ep in endpoint:
        connector.xadd(ep, {"data": 1})
    while cb_mock.call_count < 2:
        time.sleep(0.1)
    assert mock.call({"data": 1}, a=1) in cb_mock.mock_calls
    cb_mock.reset_mock()

    for ep in endpoint:
        connector.xadd(ep, {"data": 2})
    while cb_mock.call_count < 2:
        time.sleep(0.1)
    assert mock.call({"data": 2}, a=1) in cb_mock.mock_calls

    num_threads = threading.active_count()
    connector.unregister(endpoint, cb=cb_mock)
    # each direct connection (newest_only) makes 1 thread
    assert threading.active_count() == num_threads - len(endpoint)


def test_register_raises_if_no_cb(connected_connector):
    connector = connected_connector
    with pytest.raises(ValueError):
        connector.register("test", cb=None, start_thread=False)


def test_register_stream_raises_if_no_cb(connected_connector):
    connector = connected_connector
    with pytest.raises(ValueError):
        connector.register(TestStreamEndpoint, cb=None, start_thread=False)


def test_register_stream_raises_if_no_topic_nor_pattern(connected_connector):
    connector = connected_connector
    with pytest.raises(ValueError):
        connector.register(None, cb=mock.Mock(), start_thread=False)


@pytest.mark.parametrize("topics", [5, [5, "test"], [5, 5], [None, "test"], None, [None, None]])
def test_register_stream_raises_if_topic_is_not_str_nor_list(connected_connector, topics):
    connector = connected_connector
    with pytest.raises(ValueError):
        connector.register(topics, cb=lambda *args, **kwargs: ..., start_thread=False)


@pytest.mark.parametrize("val", [messages.ScanMessage(point_id=5, scan_id="1234", data={"a": 1})])
def test_lrange_parses_messages(connected_connector, val):
    connected_connector.lpush("test", val)
    res = connected_connector.lrange("test", 0, -1)
    assert isinstance(res[0], messages.ScanMessage)


def test_redis_connector_message_alternated_pass(connected_connector):
    data_original = {
        "mca1": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        "mca2": [10, 9, 8, 7, 6, 5, 4, 3, 2, 1],
    }

    data_to_check = {
        "mca1": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        "mca2": [10, 9, 8, 7, 6, 5, 4, 3, 2, 1],
    }  # same as data_original, but not passed through xadd method

    assert connected_connector.xread(topic="topic") == None
    connected_connector.xadd("topic", data_original)
    msg_received = connected_connector.xread(topic="topic")

    assert msg_received == [data_to_check]
    assert msg_received == [data_original]



"""
This module provides a mixin class for the BEC class that allows the user to load and unload scripts from the `scripts` directory.
"""

from __future__ import annotations

import builtins
import glob
import importlib
import inspect
import os
import pathlib
from typing import TYPE_CHECKING

from pylint import lint
from pylint.reporters import CollectingReporter
from rich.console import Console
from rich.table import Table

from bec_lib.callback_handler import EventType
from bec_lib.logger import bec_logger

if TYPE_CHECKING:
    from pylint.message import Message

logger = bec_logger.logger

try:
    from bec_plugins import bec_ipython_client as client_plugins
except ImportError:
    client_plugins = None


class UserScriptsMixin:
    def __init__(self) -> None:
        super().__init__()
        self._scripts = {}

    def load_all_user_scripts(self) -> None:
        """Load all scripts from the `scripts` directory.

        Runs a callback of type `EventType.NAMESPACE_UPDATE`
        to inform clients about added objects in the namesapce.
        """
        self.forget_all_user_scripts()

        # load all scripts from the scripts directory
        current_path = pathlib.Path(__file__).parent.resolve()
        script_files = glob.glob(os.path.abspath(os.path.join(current_path, "../scripts/*.py")))

        # load all scripts from the user's script directory in the home directory
        user_script_dir = os.path.join(os.path.expanduser("~"), "bec", "scripts")
        if os.path.exists(user_script_dir):
            script_files.extend(glob.glob(os.path.abspath(os.path.join(user_script_dir, "*.py"))))

        # load scripts from the plugins
        if client_plugins:
            plugin_scripts_dir = os.path.join(client_plugins.__path__[0], "scripts")
            if os.path.exists(plugin_scripts_dir):
                script_files.extend(
                    glob.glob(os.path.abspath(os.path.join(plugin_scripts_dir, "*.py")))
                )

        for file in script_files:
            self.load_user_script(file)
        builtins.__dict__.update({name: v["cls"] for name, v in self._scripts.items()})

    def forget_all_user_scripts(self) -> None:
        """unload / remove loaded user scripts from builtins. Files will remain untouched.

        Runs a callback of type `EventType.NAMESPACE_UPDATE`
        to inform clients about removing objects from the namesapce.

        """
        for name, obj in self._scripts.items():
            builtins.__dict__.pop(name)
            self.callbacks.run(
                EventType.NAMESPACE_UPDATE, action="remove", ns_objects={name: obj["cls"]}
            )
        self._scripts.clear()

    def load_user_script(self, file: str) -> None:
        """load a user script file and import all its definitions

        Args:
            file (str): Full path to the script file.
        """
        # TODO: re-enable linter
        # self._run_linter_on_file(file)
        module_members = self._load_script_module(file)
        for name, cls in module_members:
            if not callable(cls):
                continue
            # ignore imported classes
            if cls.__module__ != "scripts":
                continue
            if name in self._scripts:
                logger.warning(f"Conflicting definitions for {name}.")
            logger.info(f"Importing {name}")
            self._scripts[name] = {"cls": cls, "fname": file}
            self.callbacks.run(EventType.NAMESPACE_UPDATE, action="add", ns_objects={name: cls})

    def forget_user_script(self, name: str) -> None:
        """unload / remove a user scripts. The file will remain on disk."""
        if name not in self._scripts:
            logger.error(f"{name} is not a known user script.")
            return
        self.callbacks.run(
            EventType.NAMESPACE_UPDATE,
            action="remove",
            ns_objects={name: self._scripts[name]["cls"]},
        )
        builtins.__dict__.pop(name)
        self._scripts.pop(name)

    def list_user_scripts(self):
        """display all currently loaded user functions"""
        console = Console()
        table = Table(title="User scripts")
        table.add_column("Name", justify="center")
        table.add_column("Location", justify="center", overflow="fold")

        for name, content in self._scripts.items():
            table.add_row(name, content.get("fname"))
        console.print(table)

    def _load_script_module(self, file) -> list:
        module_spec = importlib.util.spec_from_file_location("scripts", file)
        plugin_module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(plugin_module)
        module_members = inspect.getmembers(plugin_module)
        return module_members

    def _run_linter_on_file(self, file) -> None:
        accepted_vars = ",".join([key for key in builtins.__dict__ if not key.startswith("_")])
        reporter = CollectingReporter()
        print(f"{accepted_vars}")
        lint.Run(
            [file, "--errors-only", f"--additional-builtins={accepted_vars}"],
            exit=False,
            reporter=reporter,
        )
        if not reporter.messages:
            return

        def _format_pylint_output(msg: Message):
            return f"Line {msg.line}, column {msg.column}: {msg.msg}."

        for msg in reporter.messages:
            logger.error(
                f"During the import of {file}, the following error was detected: \n{_format_pylint_output(msg)}.\nThe script was imported but may not work as expected."
            )



from bec_lib import BECService, BECStatus, ServiceConfig
from bec_lib.connector import ConnectorBase
from bec_server.scihub.scibec import SciBecConnector
from bec_server.scihub.scilog import SciLogConnector


class SciHub(BECService):
    def __init__(self, config: ServiceConfig, connector_cls: ConnectorBase) -> None:
        super().__init__(config, connector_cls, unique_service=True)
        self.config = config
        self.scibec_connector = None
        self.scilog_connector = None
        self._start_scibec_connector()
        self._start_scilog_connector()
        self.status = BECStatus.RUNNING

    def _start_scibec_connector(self):
        self.wait_for_service("DeviceServer")
        self.scibec_connector = SciBecConnector(self, self.connector)

    def _start_scilog_connector(self):
        self.scilog_connector = SciLogConnector(self, self.connector)

    def shutdown(self):
        super().shutdown()
        if self.scibec_connector:
            self.scibec_connector.shutdown()
        if self.scilog_connector:
            self.scilog_connector.shutdown()



from __future__ import annotations

import os
import time
import traceback
import uuid
from typing import TYPE_CHECKING

import bec_lib
from bec_lib import DeviceConfigError
from bec_lib import DeviceManagerBase as DeviceManager
from bec_lib import MessageEndpoints, bec_logger, messages
from bec_lib.connector import ConnectorBase
from bec_lib.scibec_validator import SciBecValidator

if TYPE_CHECKING:
    from bec_lib.device import DeviceBase
    from bec_server.scihub.scibec.scibec_connector import SciBecConnector

logger = bec_logger.logger

dir_path = os.path.abspath(os.path.join(os.path.dirname(bec_lib.__file__), "./configs/"))


class ConfigHandler:
    def __init__(self, scibec_connector: SciBecConnector, connector: ConnectorBase) -> None:
        self.scibec_connector = scibec_connector
        self.connector = connector
        self.device_manager = DeviceManager(self.scibec_connector.scihub)
        self.device_manager.initialize(scibec_connector.config.redis)
        self.validator = SciBecValidator(os.path.join(dir_path, "openapi_schema.json"))

    def parse_config_request(self, msg: messages.DeviceConfigMessage) -> None:
        """Processes a config request. If successful, it emits a config reply

        Args:
            msg (BMessage.DeviceConfigMessage): Config request

        """
        try:
            self.device_manager.check_request_validity(msg)
            if msg.content["action"] == "update":
                self._update_config(msg)
            if msg.content["action"] == "reload":
                self._reload_config(msg)
            if msg.content["action"] == "set":
                self._set_config(msg)

        except Exception:
            content = traceback.format_exc()
            self.send_config_request_reply(accepted=False, error_msg=content, metadata=msg.metadata)

    def send_config(self, msg: messages.DeviceConfigMessage) -> None:
        """broadcast a new config"""
        self.connector.send(MessageEndpoints.device_config_update(), msg)

    def send_config_request_reply(self, accepted, error_msg, metadata):
        """send a config request reply"""
        msg = messages.RequestResponseMessage(
            accepted=accepted, message=error_msg, metadata=metadata
        )
        RID = metadata.get("RID")
        self.connector.set(MessageEndpoints.device_config_request_response(RID), msg, expire=60)

    def _set_config(self, msg: messages.DeviceConfigMessage):
        config = msg.content["config"]
        scibec = self.scibec_connector.scibec
        logger.debug(self.scibec_connector.scibec_info)
        experiment = self.scibec_connector.scibec_info.get("beamline", {}).get("activeExperiment")

        msg.metadata["updated_config"] = False
        for name, device in config.items():
            self._convert_to_db_config(name, device)
            self.validator.validate_device(device)
        self.scibec_connector.set_redis_config(list(config.values()))
        msg.metadata["updated_config"] = True
        RID = str(uuid.uuid4())
        self._update_device_server(RID, config, action="reload")
        accepted, server_response_msg = self._wait_for_device_server_update(RID, timeout_time=20)
        if "failed_devices" in server_response_msg.metadata:
            msg.metadata["failed_devices"] = server_response_msg.metadata["failed_devices"]
        reload_msg = messages.DeviceConfigMessage(action="reload", config={}, metadata=msg.metadata)
        if accepted:
            self.send_config_request_reply(accepted=accepted, error_msg=None, metadata=msg.metadata)
            self.send_config(reload_msg)
            return
        self.send_config_request_reply(
            accepted=accepted,
            error_msg=f"{server_response_msg.message} Error during loading. The config will be flushed",
            metadata=msg.metadata,
        )
        self.send_config(reload_msg)

    def _convert_to_db_config(self, name: str, config: dict) -> None:
        if "deviceConfig" in config and config["deviceConfig"] is None:
            config["deviceConfig"] = {}
        config["name"] = name

    def _reload_config(self, msg: messages.DeviceConfigMessage):
        # if we have a connection to SciBec, pull the data before forwarding the reload request
        if self.scibec_connector.scibec:
            self.scibec_connector.update_session()
        self.send_config_request_reply(accepted=True, error_msg=None, metadata=msg.metadata)
        self.send_config(msg)
        # self.device_manager.update_status(BECStatus.BUSY)
        # self.device_manager.devices.flush()
        # self.device_manager._get_config()
        # self.device_manager.update_status(BECStatus.RUNNING)

    def _update_config(self, msg: messages.DeviceConfigMessage):
        updated = False
        dev_configs = msg.content["config"]

        for dev, config in dev_configs.items():
            device = self.device_manager.devices[dev]
            updated = self._update_device_config(device, config.copy())
            if updated:
                if "id" in device._config and self.scibec_connector.scibec:
                    self.scibec_connector.scibec.patch_device_config(device._config["id"], config)
                self.update_config_in_redis(device)

        # send updates to services
        if updated:
            self.send_config(msg)
            self.send_config_request_reply(accepted=True, error_msg=None, metadata=msg.metadata)

    def _update_device_server(self, RID: str, config: dict, action="update") -> None:
        msg = messages.DeviceConfigMessage(action=action, config=config, metadata={"RID": RID})
        self.connector.send(MessageEndpoints.device_server_config_request(), msg)

    def _wait_for_device_server_update(self, RID: str, timeout_time=10) -> bool:
        timeout = timeout_time
        time_step = 0.05
        elapsed_time = 0
        while True:
            msg = self.connector.get(MessageEndpoints.device_config_request_response(RID))
            if msg:
                return msg.content["accepted"], msg

            if elapsed_time > timeout:
                raise TimeoutError(
                    "Reached timeout whilst waiting for a device server config reply."
                )

            time.sleep(time_step)
            elapsed_time += time_step

    def _update_device_config(self, device: DeviceBase, dev_config) -> bool:
        updated = False
        if "deviceConfig" in dev_config:
            RID = str(uuid.uuid4())
            self._update_device_server(RID, {device.name: dev_config})
            updated, msg = self._wait_for_device_server_update(RID)
            if not updated:
                raise DeviceConfigError(f"Failed to update device {device.name}. {msg.message}")
            device._config["deviceConfig"].update(dev_config["deviceConfig"])
            dev_config.pop("deviceConfig")

        if "enabled" in dev_config:
            self._validate_update({"enabled": dev_config["enabled"]})
            device._config["enabled"] = dev_config["enabled"]
            RID = str(uuid.uuid4())
            self._update_device_server(RID, {device.name: dev_config})
            updated, msg = self._wait_for_device_server_update(RID)
            if not updated:
                raise DeviceConfigError(f"Failed to update device {device.name}. {msg.message}")
            dev_config.pop("enabled")

        if not dev_config:
            return updated

        available_keys = [
            "readOnly",
            "userParameter",
            "onFailure",
            "deviceTags",
            "readoutPriority",
            "softwareTrigger",
        ]
        for key in dev_config:
            if key not in available_keys:
                raise DeviceConfigError(f"Unknown update key {key}.")

            self._validate_update({key: dev_config[key]})
            device._config[key] = dev_config[key]
            updated = True

        return updated

    def _validate_update(self, update):
        self.validator.validate_device_patch(update)

    def update_config_in_redis(self, device):
        config = self.device_manager.connector.get(MessageEndpoints.device_config())
        config = config.content["resource"]
        index = next(
            index for index, dev_conf in enumerate(config) if dev_conf["name"] == device.name
        )
        config[index] = device._config
        msg = messages.AvailableResourceMessage(resource=config)
        self.device_manager.connector.set(MessageEndpoints.device_config(), msg)



"""
This module provides the DAPPlugins class, which is used to access all available DAP plugins.
"""

from __future__ import annotations

import bec_lib.dap_plugin_objects as dap_plugin_objects
from bec_lib.endpoints import MessageEndpoints
from bec_lib.logger import bec_logger
from bec_lib.signature_serializer import dict_to_signature

logger = bec_logger.logger


class DAPPlugins:
    """
    DAPPlugins is a class that provides access to all available DAP plugins.
    """

    def __init__(self, parent):
        self._parent = parent
        self._available_dap_plugins = {}
        self._import_dap_plugins()
        self._selected_model = None
        self._auto_run = False
        self._selected_device = None

    def refresh(self):
        """
        Refresh the list of available DAP plugins. This is useful if new plugins have been added after
        the client has been initialized. This method is called automatically when the client is initialized.
        A call to this method is indempotent, meaning it can be called multiple times without side effects.
        """
        self._import_dap_plugins()

    def _import_dap_plugins(self):
        available_services = self._parent.service_status
        if not available_services:
            # not sure how we got here...
            return
        dap_services = [
            service for service in available_services if service.startswith("DAPServer/")
        ]
        for service in dap_services:
            available_plugins = self._parent.connector.get(
                MessageEndpoints.dap_available_plugins(service)
            )
            if available_plugins is None:
                logger.warning("No plugins available. Are redis and the BEC server running?")
                return
            for plugin_name, plugin_info in available_plugins.content["resource"].items():
                try:
                    if plugin_name in self._available_dap_plugins:
                        continue
                    name = plugin_info["user_friendly_name"]
                    auto_run_supported = plugin_info.get("auto_run_supported", False)
                    plugin_cls = self._get_plugin_class(plugin_info)
                    self._available_dap_plugins[name] = plugin_cls(
                        name,
                        plugin_info,
                        client=self._parent,
                        auto_run_supported=auto_run_supported,
                        service_info=available_services[service].content,
                    )
                    self._set_plugin(
                        name,
                        plugin_info.get("class_doc"),
                        plugin_info.get("run_doc"),
                        plugin_info.get("run_name"),
                        plugin_info.get("signature"),
                    )
                # pylint: disable=broad-except
                except Exception as e:
                    logger.error(f"Error importing plugin {plugin_name}: {e}")

    def _get_plugin_class(self, plugin_info):
        if hasattr(dap_plugin_objects, plugin_info["class"]):
            return getattr(dap_plugin_objects, plugin_info["class"])
        if plugin_info.get("auto_run_supported"):
            return dap_plugin_objects.DAPPluginObjectAutoRun
        return dap_plugin_objects.DAPPluginObject

    def _set_plugin(
        self,
        plugin_name: str,
        class_doc_string: str,
        run_doc_string: str,
        run_name: str,
        signature: dict,
    ):
        # pylint disable=protected-access
        setattr(self, plugin_name, self._available_dap_plugins[plugin_name])
        plugin = getattr(self, plugin_name)
        setattr(plugin, "__doc__", class_doc_string)
        setattr(plugin, run_name, plugin._user_run)
        setattr(plugin._user_run, "__doc__", run_doc_string)
        setattr(plugin._user_run, "__signature__", dict_to_signature(signature))



from __future__ import annotations

import threading
import time
import uuid
from collections.abc import Callable
from typing import Generator, Literal

import numpy as np

from bec_lib import MessageEndpoints, Status, bec_logger, messages
from bec_lib.connector import ConnectorBase

from .errors import DeviceMessageError, ScanAbortion

logger = bec_logger.logger


class ScanStubs:
    def __init__(
        self,
        connector: ConnectorBase,
        device_msg_callback: Callable = None,
        shutdown_event: threading.Event = None,
    ) -> None:
        self.connector = connector
        self.device_msg_metadata = (
            device_msg_callback if device_msg_callback is not None else lambda: {}
        )
        self.shutdown_event = shutdown_event

    @staticmethod
    def _exclude_nones(input_dict: dict):
        for key in list(input_dict.keys()):
            if input_dict[key] is None:
                input_dict.pop(key)

    def _device_msg(self, **kwargs):
        """"""
        msg = messages.DeviceInstructionMessage(**kwargs)
        msg.metadata = {**self.device_msg_metadata(), **msg.metadata}
        return msg

    def send_rpc_and_wait(self, device: str, func_name: str, *args, **kwargs) -> any:
        """Perform an RPC (remote procedure call) on a device and wait for its return value.

        Args:
            device (str): Name of the device
            func_name (str): Function name. The function name will be appended to the device.
            args (tuple): Arguments to pass on to the RPC function
            kwargs (dict): Keyword arguments to pass on to the RPC function

        Raises:
            ScanAbortion: Raised if the RPC's success is False

        Returns:
            any: Return value of the executed rpc function

        Examples:
            >>> send_rpc_and_wait("samx", "controller.my_custom_function")
        """
        rpc_id = str(uuid.uuid4())
        parameter = {
            "device": device,
            "func": func_name,
            "rpc_id": rpc_id,
            "args": args,
            "kwargs": kwargs,
        }
        yield from self.rpc(device=device, parameter=parameter, metadata={"response": True})
        return self._get_from_rpc(rpc_id)

    def _get_from_rpc(self, rpc_id) -> any:
        """
        Get the return value from an RPC call.

        Args:
            rpc_id (str): RPC ID

        Raises:
            ScanAbortion: Raised if the RPC's success flag is False

        Returns:
            any: Return value of the RPC call
        """

        while not self.shutdown_event.is_set():
            msg = self.connector.get(MessageEndpoints.device_rpc(rpc_id))
            if msg:
                break
            time.sleep(0.001)
        if self.shutdown_event.is_set():
            raise ScanAbortion("The scan was aborted.")
        if not msg.content["success"]:
            error = msg.content["out"]
            if isinstance(error, dict) and {"error", "msg", "traceback"}.issubset(
                set(error.keys())
            ):
                error_msg = f"During an RPC, the following error occured:\n{error['error']}: {error['msg']}.\nTraceback: {error['traceback']}\n The scan will be aborted."
            else:
                error_msg = "During an RPC, an error occured"
            raise ScanAbortion(error_msg)

        logger.debug(msg.content.get("out"))
        return_val = msg.content.get("return_val")
        if not isinstance(return_val, dict):
            return return_val
        if return_val.get("type") == "status" and return_val.get("RID"):
            return Status(self.connector, return_val.get("RID"))
        return return_val

    def set_and_wait(
        self, *, device: list[str], positions: list | np.ndarray
    ) -> Generator[None, None, None]:
        """Set devices to a specific position and wait completion.

        Args:
            device (list[str]): List of device names.
            positions (list | np.ndarray): Target position.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.

        """
        if not isinstance(positions, list) and not isinstance(positions, np.ndarray):
            positions = [positions]
        if len(positions) == 0:
            return
        for ind, val in enumerate(device):
            yield from self.set(device=val, value=positions[ind], wait_group="scan_motor")
        yield from self.wait(device=device, wait_type="move", wait_group="scan_motor")

    def read_and_wait(
        self, *, wait_group: str, device: list = None, group: str = None, point_id: int = None
    ) -> Generator[None, None, None]:
        """Trigger a reading and wait for completion.

        Args:
            wait_group (str): wait group
            device (list, optional): List of device names. Can be specified instead of group. Defaults to None.
            group (str, optional): Group name of devices. Can be specified instead of device. Defaults to None.
            point_id (int, optional): _description_. Defaults to None.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.

        """
        self._check_device_and_groups(device, group)
        yield from self.read(device=device, group=group, wait_group=wait_group, point_id=point_id)
        yield from self.wait(device=device, wait_type="read", group=group, wait_group=wait_group)

    def open_scan(
        self,
        *,
        scan_motors: list,
        readout_priority: dict,
        num_pos: int,
        scan_name: str,
        scan_type: Literal["step", "fly"],
        positions=None,
        metadata=None,
    ) -> Generator[None, None, None]:
        """Open a new scan.

        Args:
            scan_motors (list): List of scan motors.
            readout_priority (dict): Modification of the readout priority.
            num_pos (int): Number of positions within the scope of this scan.
            positions (list): List of positions for this scan.
            scan_name (str): Scan name.
            scan_type (str): Scan type (e.g. 'step' or 'fly')

        Returns:
            Generator[None, None, None]: Generator that yields a device message.

        """
        yield self._device_msg(
            device=None,
            action="open_scan",
            parameter={
                "scan_motors": scan_motors,
                "readout_priority": readout_priority,
                "num_points": num_pos,
                "positions": positions,
                "scan_name": scan_name,
                "scan_type": scan_type,
            },
            metadata=metadata,
        )

    def kickoff(
        self, *, device: str, parameter: dict = None, wait_group="kickoff", metadata=None
    ) -> Generator[None, None, None]:
        """Kickoff a fly scan device.

        Args:
            device (str): Device name of flyer.
            parameter (dict, optional): Additional parameters that should be forwarded to the device. Defaults to {}.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        parameter = parameter if parameter is not None else {}
        parameter = {"configure": parameter, "wait_group": wait_group}
        yield self._device_msg(
            device=device, action="kickoff", parameter=parameter, metadata=metadata
        )

    def complete(self, *, device: str, metadata=None) -> Generator[None, None, None]:
        """Complete a fly scan device.

        Args:
            device (str): Device name of flyer.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        yield self._device_msg(device=device, action="complete", parameter={}, metadata=metadata)

    def get_req_status(self, device: str, RID: str, DIID: int) -> int:
        """Check if a device request status matches the given RID and DIID

        Args:
            device (str): device under inspection
            RID (str): request ID
            DIID (int): device instruction ID

        Returns:
            int: 1 if the request status matches the RID and DIID, 0 otherwise
        """
        msg = self.connector.get(MessageEndpoints.device_req_status(device))
        if not msg:
            return 0
        matching_RID = msg.metadata.get("RID") == RID
        matching_DIID = msg.metadata.get("DIID") == DIID
        if matching_DIID and matching_RID:
            return 1
        return 0

    def get_device_progress(self, device: str, RID: str) -> float | None:
        """Get reported device progress

        Args:
            device (str): Name of the device
            RID (str): request ID

        Returns:
            float: reported progress value

        """
        msg = self.connector.get(MessageEndpoints.device_progress(device))
        if not msg:
            return None
        matching_RID = msg.metadata.get("RID") == RID
        if not matching_RID:
            return None
        if not isinstance(msg, messages.ProgressMessage):
            raise DeviceMessageError(
                f"Expected to receive a Progressmessage for device {device} but instead received {msg}."
            )
        return msg.content["value"]

    def close_scan(self) -> Generator[None, None, None]:
        """
        Close the scan.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """

        yield self._device_msg(device=None, action="close_scan", parameter={})

    def stage(self) -> Generator[None, None, None]:
        """
        Stage all devices

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        yield self._device_msg(device=None, action="stage", parameter={})

    def unstage(self) -> Generator[None, None, None]:
        """
        Unstage all devices

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        yield self._device_msg(device=None, action="unstage", parameter={})

    def pre_scan(self) -> Generator[None, None, None]:
        """
        Trigger pre-scan actions on all devices. Typically, pre-scan actions are called directly before the scan core starts and
        are used to perform time-critical actions.
        The event will be sent to all devices that have a pre_scan method implemented.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        yield self._device_msg(device=None, action="pre_scan", parameter={})

    def baseline_reading(self) -> Generator[None, None, None]:
        """
        Run the baseline readings. This will readout all devices that are marked with the readout_priority "baseline".

        Returns:
            Generator[None, None, None]: Generator that yields a device message.

        """
        yield self._device_msg(
            device=None,
            action="baseline_reading",
            parameter={},
            metadata={"readout_priority": "baseline"},
        )

    def wait(
        self,
        *,
        wait_type: Literal["move", "read", "trigger"],
        device: list[str] | str | None = None,
        group: Literal["scan_motor", "primary", None] = None,
        wait_group: str = None,
        wait_time: float = None,
    ):
        """Wait for an event.

        Args:
            wait_type (Literal["move", "read", "trigger"]): Type of wait event. Can be "move", "read" or "trigger".
            device (list[str] | str, optional): List of device names. Defaults to None.
            group (Literal["scan_motor", "primary", None]): Device group that can be used instead of device. Defaults to None.
            wait_group (str, optional): Wait group for this event. Defaults to None.
            wait_time (float, optional): Wait time (for wait_type="trigger"). Defaults to None.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.

        Example:
            >>> yield from self.stubs.wait(wait_type="move", group="scan_motor", wait_group="scan_motor")
            >>> yield from self.stubs.wait(wait_type="read", group="scan_motor", wait_group="my_readout_motors")

        """
        self._check_device_and_groups(device, group)
        parameter = {"type": wait_type, "time": wait_time, "group": group, "wait_group": wait_group}
        self._exclude_nones(parameter)
        yield self._device_msg(device=device, action="wait", parameter=parameter)

    def read(
        self,
        *,
        wait_group: str,
        device: list[str] | str | None = None,
        point_id: int | None = None,
        group: Literal["scan_motor", "primary", None] = None,
    ) -> Generator[None, None, None]:
        """
        Trigger a reading on a device or device group.

        Args:
            wait_group (str): Wait group for this event. The specified wait group can later be used
                to wait for the completion of this event. Please note that the wait group has to be
                unique. within the scope of the read / wait event.
            device (list, optional): Device name. Can be used instead of group. Defaults to None.
            point_id (int, optional): point_id to assign this reading to point within the scan. Defaults to None.
            group (Literal["scan_motor", "primary", None], optional): Device group. Can be used instead of device. Defaults to None.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.

        Example:
            >>> yield from self.stubs.read(wait_group="readout_primary", group="primary", point_id=self.point_id)
            >>> yield from self.stubs.read(wait_group="sample_stage", device="samx", point_id=self.point_id)

        """
        self._check_device_and_groups(device, group)
        parameter = {"group": group, "wait_group": wait_group}
        metadata = {"point_id": point_id}
        self._exclude_nones(parameter)
        self._exclude_nones(metadata)
        yield self._device_msg(device=device, action="read", parameter=parameter, metadata=metadata)

    def publish_data_as_read(
        self, *, device: str, data: dict, point_id: int
    ) -> Generator[None, None, None]:
        """
        Publish the given data as a read event and assign it to the given point_id.
        This method can be used to customize the assignment of data to a specific point within a scan.

        Args:
            device (str): Device name.
            data (dict): Data that should be published.
            point_id (int): point_id that should be attached to this data.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        metadata = {"point_id": point_id}
        yield self._device_msg(
            device=device,
            action="publish_data_as_read",
            parameter={"data": data},
            metadata=metadata,
        )

    def trigger(self, *, group: str, point_id: int) -> Generator[None, None, None]:
        """Trigger a device group

        Args:
            group (str): Device group that should receive the trigger.
            point_id (int): point_id that should be attached to this trigger event.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        yield self._device_msg(
            device=None,
            action="trigger",
            parameter={"group": group},
            metadata={"point_id": point_id},
        )

    def set(self, *, device: str, value: float, wait_group: str, metadata=None):
        """Set the device to a specific value.

        Args:
            device (str): Device name
            value (float): Target value.
            wait_group (str): wait group for this event.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.

        """
        yield self._device_msg(
            device=device,
            action="set",
            parameter={"value": value, "wait_group": wait_group},
            metadata=metadata,
        )

    def open_scan_def(self) -> Generator[None, None, None]:
        """
        Open a new scan definition

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        yield self._device_msg(device=None, action="open_scan_def", parameter={})

    def close_scan_def(self) -> Generator[None, None, None]:
        """
        Close a scan definition

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        yield self._device_msg(device=None, action="close_scan_def", parameter={})

    def close_scan_group(self) -> Generator[None, None, None]:
        """
        Close a scan group

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        yield self._device_msg(device=None, action="close_scan_group", parameter={})

    def rpc(self, *, device: str, parameter: dict, metadata=None) -> Generator[None, None, None]:
        """Perfrom an RPC (remote procedure call) on a device.

        Args:
            device (str): Device name.
            parameter (dict): parameters used for this rpc instructions.

        Returns:
            Generator[None, None, None]: Generator that yields a device message.

        """
        yield self._device_msg(device=device, action="rpc", parameter=parameter, metadata=metadata)

    def scan_report_instruction(self, instructions: dict) -> Generator[None, None, None]:
        """Scan report instructions

        Args:
            instructions (dict): Dict containing the scan report instructions

        Returns:
            Generator[None, None, None]: Generator that yields a device message.
        """
        yield self._device_msg(
            device=None, action="scan_report_instruction", parameter=instructions
        )

    def _check_device_and_groups(self, device, group) -> None:
        if device and group:
            raise DeviceMessageError("Device and device group was specified. Pick one.")
        if device is None and group is None:
            raise DeviceMessageError("Either devices or device groups have to be specified.")



import inspect
from typing import Literal, Union

import pytest

from bec_lib.signature_serializer import (
    deserialize_dtype,
    dict_to_signature,
    serialize_dtype,
    signature_to_dict,
)


def test_signature_serializer():
    def test_func(a, b, c=1, d=2, e: int = 3):
        pass

    params = signature_to_dict(test_func)
    assert params == [
        {"name": "a", "kind": "POSITIONAL_OR_KEYWORD", "default": "_empty", "annotation": "_empty"},
        {"name": "b", "kind": "POSITIONAL_OR_KEYWORD", "default": "_empty", "annotation": "_empty"},
        {"name": "c", "kind": "POSITIONAL_OR_KEYWORD", "default": 1, "annotation": "_empty"},
        {"name": "d", "kind": "POSITIONAL_OR_KEYWORD", "default": 2, "annotation": "_empty"},
        {"name": "e", "kind": "POSITIONAL_OR_KEYWORD", "default": 3, "annotation": "int"},
    ]

    sig = dict_to_signature(params)
    assert sig == inspect.signature(test_func)


def test_signature_serializer_with_literals():
    def test_func(a, b: Literal[1, 2, 3] = 1):
        pass

    params = signature_to_dict(test_func)
    assert params == [
        {"name": "a", "kind": "POSITIONAL_OR_KEYWORD", "default": "_empty", "annotation": "_empty"},
        {
            "name": "b",
            "kind": "POSITIONAL_OR_KEYWORD",
            "default": 1,
            "annotation": {"Literal": (1, 2, 3)},
        },
    ]

    sig = dict_to_signature(params)
    assert sig == inspect.signature(test_func)


@pytest.mark.parametrize(
    "dtype_in,dtype_out",
    [
        (int, "int"),
        (str, "str"),
        (float, "float"),
        (bool, "bool"),
        (inspect._empty, "_empty"),
        # (Literal[1, 2, 3], {"Literal": (1, 2, 3)}),
        # (Union[int, str], "int | str"),
    ],
)
def test_serialize_dtype(dtype_in, dtype_out):
    assert dtype_out == serialize_dtype(dtype_in)


@pytest.mark.parametrize(
    "dtype_in,dtype_out",
    [
        ("int", int),
        ("str", str),
        ("float", float),
        ("bool", bool),
        ("_empty", inspect._empty),
        ({"Literal": (1, 2, 3)}, Literal[1, 2, 3]),
        ("int | str", Union[int, str]),
    ],
)
def test_deserialize_dtype(dtype_in, dtype_out):
    assert dtype_out == deserialize_dtype(dtype_in)



from unittest import mock

import pytest

from bec_lib import MessageEndpoints, messages
from bec_server.scan_bundler.bec_emitter import BECEmitter


@pytest.fixture
def bec_emitter_mock(scan_bundler_mock):
    emitter = BECEmitter(scan_bundler_mock)
    yield emitter
    emitter.shutdown()


def test_on_scan_point_emit_BEC(bec_emitter_mock):
    sb = bec_emitter_mock.scan_bundler
    with mock.patch.object(bec_emitter_mock, "_send_bec_scan_point") as send:
        bec_emitter_mock.on_scan_point_emit("scan_id", 2)
        send.assert_called_once_with("scan_id", 2)


def test_on_baseline_emit_BEC(bec_emitter_mock):
    sb = bec_emitter_mock.scan_bundler
    with mock.patch.object(bec_emitter_mock, "_send_baseline") as send:
        bec_emitter_mock.on_baseline_emit("scan_id")
        send.assert_called_once_with("scan_id")


def test_send_bec_scan_point(bec_emitter_mock):
    sb = bec_emitter_mock.scan_bundler
    scan_id = "lkajsdlkj"
    point_id = 2
    sb.sync_storage[scan_id] = {"info": {}, "status": "open", "sent": set()}
    sb.sync_storage[scan_id][point_id] = {}
    msg = messages.ScanMessage(
        point_id=point_id,
        scan_id=scan_id,
        data=sb.sync_storage[scan_id][point_id],
        metadata={"scan_id": "lkajsdlkj", "scan_type": None, "scan_report_devices": None},
    )
    with mock.patch.object(bec_emitter_mock, "add_message") as send:
        bec_emitter_mock._send_bec_scan_point(scan_id, point_id)
        send.assert_called_once_with(
            msg,
            MessageEndpoints.scan_segment(),
            MessageEndpoints.public_scan_segment(scan_id, point_id),
        )


def test_send_baseline_BEC(bec_emitter_mock):
    sb = bec_emitter_mock.scan_bundler
    scan_id = "lkajsdlkj"
    sb.sync_storage[scan_id] = {"info": {}, "status": "open", "sent": set()}
    sb.sync_storage[scan_id]["baseline"] = {}
    msg = messages.ScanBaselineMessage(scan_id=scan_id, data=sb.sync_storage[scan_id]["baseline"])
    with mock.patch.object(sb, "connector") as connector:
        bec_emitter_mock._send_baseline(scan_id)
        pipe = connector.pipeline()
        connector.set.assert_called_once_with(
            MessageEndpoints.public_scan_baseline(scan_id), msg, expire=1800, pipe=pipe
        )
        connector.set_and_publish.assert_called_once_with(
            MessageEndpoints.scan_baseline(), msg, pipe=pipe
        )
